/*
 *  File: rvp.c - Gaim RVP plug-in implementation.
 *  Copyright (C) 2003  Weihua Sun (weihua@lucent.com)
 *  Updated 2005 Waider <waider@waider.ie>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#define LOUD /* for debugging loudly */
#define GAIM_PLUGINS /* this switches the GAIM_INIT_PLUGIN declaration
                        from in-line build to outboard build */

#include "rvp.h"

#include <netdb.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <iconv.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <arpa/nameser_compat.h>
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "random.h"

/* gaim includes */
#include <accountopt.h>
#include <debug.h>
#include <internal.h> /* this is here for the gettext bits */
#include <md5.h>
#include <notify.h>
#include <plugin.h>
#include <pluginpref.h>
#include <privacy.h>
#include <util.h>
#include <version.h>

#include <libxml/debugXML.h>

static char HEX_ELM[17] = { '0','1','2','3','4','5','6','7',
                            '8','9','A','B','C','D','E','F','\0' };

static void rvp_send_request( GaimConnection *, gchar *, gchar ** );
static void rvp_add_buddy( GaimConnection *, GaimBuddy *, GaimGroup * );
static void rvp_set_buddy_status( GaimConnection *, gchar *, char * );
static const char *rvp_normalize( const GaimAccount *, const char * );
static void rvp_send_request( GaimConnection *, gchar *, gchar ** );

/* async url stuff */
static void rvp_async_data( void *, GaimFetchUrlData *, size_t );
static void url_fetched_cb( gpointer, gint, GaimInputCondition );
static void destroy_fetch_url_data( GaimFetchUrlData * );

static void rvp_add_permit(GaimConnection *, const char *);

void rvp_chat_user_left( GaimConnection *, int, GaimBuddy * );

/*
 * This horror is just a temporary hack until I get around to redoing
 * it properly... right now I just want the SRV record ASAP.
 *
 * correct version:
 * 1. asynchronous, if possible. see the main gaim source
 *    1a. need to handle callbacks for this, which complicates code
 *        elsewhere
 * 2. need to keep track of use of SRV records and balance them
 *    accordingly. This is the intention of the unused 'last' param.
 * 3. cache what we've learned and short circuit responses. NB TTL!
 */
/* record type: 2 bytes
   record class: 2 bytes
   TTL: 4 bytes
   data length: 2 bytes */
typedef struct _rechdr {
  guint16 type;
  guint16 class;
  guint32 ttl;
  guint16 length;
} rechdr;

/* cache for SRV records */
static GHashTable *srvcache = NULL;

srvrec *gethostbysrv( char *srvname, srvrec *last ) {
  HEADER *hdr;
  unsigned char answer[1024];
  int len = 0;
  char name[1024];
  srvrec *retval = NULL;

  if ( srvcache == NULL ) {
    gaim_debug_misc( __FUNCTION__, "Created SRV cache\n" );
    srvcache = g_hash_table_new( g_str_hash, g_str_equal );
  }

  if (( retval = (srvrec *)g_hash_table_lookup( srvcache, srvname )) != NULL ) {
    if ( retval->expiry > time( NULL )) {
      return retval;
    } else {
      gaim_debug_misc( __FUNCTION__, "cached %s expired\n", srvname );
      retval = NULL;
    }
  } else {
    gaim_debug_misc( __FUNCTION__, "%s not in SRV cache\n", srvname );
  }

  bzero( answer, 1024 );
  bzero( name, 1024 );

  _res.options |= RES_DEBUG;
  if (( len = res_search( srvname, C_IN, T_SRV, answer, sizeof( answer ))) != -1 ) {
    unsigned char *blob;
    int i, l;

    hdr = (HEADER *)answer;
    if ( ntohs( hdr->ancount ) == 0 ) {
      gaim_debug_warning( __FUNCTION__, "res_search: no records found\n" );
      goto out;
    }

    blob = &answer[HFIXEDSZ];

    for ( i = 0; i < ntohs( hdr->qdcount ); i++ ) {
      l = dn_expand( answer, answer + len, blob, name, 1024 );
      if ( l < 0 ) {
        gaim_debug_error( __FUNCTION__, "dn_expand failed (1)\n" );
        goto out;
      }
      blob += l + QFIXEDSZ;
    }

    for ( i = 0; i < ntohs( hdr->ancount ); i++ ) {
      rechdr header;

      l = dn_expand( answer, answer + len, blob, name, 1024 );
      if ( l < 0 ) {
        gaim_debug_error( __FUNCTION__, "dn_expand failed (2)\n" );
        return NULL;
      }

      blob += l;

      memcpy( &header, blob, sizeof( header ));

      /* data:
         priority: 2 bytes
         weight: 2 bytes
         port: 2 bytes
         target: remainder
      */
      l = dn_expand( answer, answer + len, blob + 16, name, 1024 );
      if ( l < 0 ) {
        gaim_debug_error( __FUNCTION__, "dn_expand failed (3)\n" );
        return NULL;
      }

      retval = g_new0( srvrec, 1 );
      retval->host = g_strdup( name );
      retval->port = ntohs( blob[14] + 256 * blob[15] );
      retval->expiry = time( NULL ) + ntohl( header.ttl );
    }
  } else {
    gaim_debug_error( __FUNCTION__, "res_search for %s failed (%s)\n",
                      srvname, strerror( errno ));
  }

 out:
  if ( retval == NULL ) {
    retval = g_new0( srvrec, 1 );
    retval->host = NULL;
    retval->port = 0;
    retval->expiry = time( NULL ) + 86400; /* xxx */
  }

  g_hash_table_replace( srvcache, g_strdup( srvname ), retval );

  return retval;
}
/* end of horror */

/*
 * debug function to dump out a buddy structure
 */
static void rvp_dump_buddy( GaimBuddy *buddy ) {
  if ( buddy != NULL ) {
    RVPBuddy *rbud = buddy->proto_data;

    gaim_debug_misc( __FUNCTION__,     "ptr       %p (rvp @ %p)\n", buddy,
                     rbud );
    gaim_debug_misc( __FUNCTION__,     "buddy     %s\n", buddy->name );
    if ( buddy->alias != NULL )
      gaim_debug_misc( __FUNCTION__,   "alias     %s\n", buddy->alias );
    if ( buddy->server_alias != NULL )
      gaim_debug_misc( __FUNCTION__,   "srvalias  %s\n", buddy->server_alias );

    if ( rbud != NULL ) {
      if ( rbud->email != NULL )
        gaim_debug_misc( __FUNCTION__, "email     %s\n", rbud->email );
      if ( rbud->principal != NULL )
        gaim_debug_misc( __FUNCTION__, "principal %s\n", rbud->principal );
      if ( rbud->sessionid != NULL )
        gaim_debug_misc( __FUNCTION__, "sessionid %s\n", rbud->sessionid );
      if ( rbud->subs_id )
        gaim_debug_misc( __FUNCTION__, "subs id   %d\n", rbud->subs_id );

      if ( rbud->acl.credentials.assertion )
        gaim_debug_misc( __FUNCTION__, "CRED: assertion\n" );
      if ( rbud->acl.credentials.digest )
        gaim_debug_misc( __FUNCTION__, "CRED: digest\n" );
      if ( rbud->acl.credentials.ntlm )
        gaim_debug_misc( __FUNCTION__, "CRED: ntlm\n" );

      if ( rbud->acl.list )
        gaim_debug_misc( __FUNCTION__, "ACL: list\n" );
      if ( rbud->acl.read )
        gaim_debug_misc( __FUNCTION__, "ACL: read\n" );
      if ( rbud->acl.write )
        gaim_debug_misc( __FUNCTION__, "ACL: write\n" );
      if ( rbud->acl.send_to )
        gaim_debug_misc( __FUNCTION__, "ACL: send-to\n" );
      if ( rbud->acl.receive_from )
        gaim_debug_misc( __FUNCTION__, "ACL: receive-from\n" );
      if ( rbud->acl.readacl )
        gaim_debug_misc( __FUNCTION__, "ACL: readacl\n" );
      if ( rbud->acl.writeacl )
        gaim_debug_misc( __FUNCTION__, "ACL: writeacl\n" );
      if ( rbud->acl.presence )
        gaim_debug_misc( __FUNCTION__, "ACL: presence\n" );
      if ( rbud->acl.subscriptions )
        gaim_debug_misc( __FUNCTION__, "ACL: subscriptions\n" );
      if ( rbud->acl.subscribe_others )
        gaim_debug_misc( __FUNCTION__, "ACL: subscribe_others\n" );
    }
  }
}

static void rvp_free_buddy( GaimBuddy *buddy ) {
  RVPBuddy *rb = buddy->proto_data;

  if ( rb->principal != NULL ) {
    g_free( rb->principal );
  }

  g_free( rb );

  /* free GaimBuddy */
}

/*
 * keep track of conversations with people who aren't on the buddy list
 */
static GaimBuddy *rvp_get_temp_buddy1( GaimConnection *gc,
                                       GaimBuddy *buddy,
                                       gboolean createp ) {
  RVPData *rd = gc->proto_data;
  RVPBuddy *rb = buddy->proto_data;
  GaimBuddy *retval = NULL;

  if ( rd->nonbuddy == NULL ) {
    gaim_debug_misc( __FUNCTION__, "creating hash table\n" );
    rd->nonbuddy = g_hash_table_new( g_str_hash, g_str_equal );
  }

  /* we may have to populate the nick field if I've been lazy */
  if ( buddy->name == NULL ) {
    gchar *nick = rindex( rb->principal, '/' );
    if ( nick != NULL ) {
      buddy->name = g_strdup( &nick[1] );
    }
  }

  /* these are edge cases which I should really catch elsewhere */
  if ( rb->buddy == NULL ) {
    rb->buddy = buddy;
  }

  if ( buddy->proto_data == NULL ) {
    buddy->proto_data = rb;
  }

  gaim_debug_misc( __FUNCTION__, "looking up %s in hash\n", rb->principal );

  retval = (GaimBuddy *)g_hash_table_lookup( rd->nonbuddy, rb->principal );
  if ( retval == NULL && createp == TRUE ) {
    gaim_debug_misc( __FUNCTION__, "didn't find %s, creating\n",
                     rb->principal );
    g_hash_table_insert( rd->nonbuddy, g_strdup( rb->principal ), buddy );
    retval = buddy;
  }

  return retval;
}

static GaimBuddy *rvp_get_temp_buddy_create( GaimConnection *gc, GaimBuddy *buddy ) {
  GaimBuddy *retval = rvp_get_temp_buddy1( gc, buddy, TRUE );

  if ( retval != NULL ) { /* and it shouldn't be */
  } else {
    gaim_debug_error( __FUNCTION__, "Doh. NULL buddy created?\n" );
  }

  return retval;
}

static GaimBuddy *rvp_get_temp_buddy( GaimConnection *gc, GaimBuddy *buddy ) {
  return rvp_get_temp_buddy1( gc, buddy, FALSE );
}

/*
 * parse an XML fragment identifying a contact or contacts
 */
static GaimBuddy **parse_contact( xmlDocPtr doc, xmlNodePtr from ) {
  xmlNodePtr contact = from;
  GaimBuddy **blist = NULL;
  gint bcount = 0;

  while( contact ) {
    xmlNodePtr ptr = contact->xmlChildrenNode;
    GaimBuddy *buddy = NULL;

    while( ptr ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "href" )) {
        xmlChar *href = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        if ( href != NULL ) {
          if ( buddy == NULL ) {
            buddy = g_new0( GaimBuddy, 1 );
          }
          buddy->proto_data = g_new0( RVPBuddy, 1 );
          ((RVPBuddy *)buddy->proto_data)->buddy = buddy;
          ((RVPBuddy *)buddy->proto_data)->principal = g_strdup( (gchar *)href );
          xmlFree( href );
        }
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "description")) {
        xmlChar *desc = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        if ( desc != NULL ) {
          if ( buddy == NULL ) {
            buddy = g_new0( GaimBuddy, 1 );
          }
          buddy->server_alias = g_strdup((gchar *)desc );
          xmlFree( desc );
        }
      }

      ptr = ptr->next;
    }

    blist = g_realloc( blist, sizeof( GaimBuddy *) * bcount + 1 );
    blist[bcount] = buddy;
    bcount++;

    contact = contact->next;
  }

  /* null terminate */
  blist = g_realloc( blist, sizeof( GaimBuddy *) * bcount + 1 );
  blist[bcount] = NULL;

  if ( bcount > 1 ) {
    gaim_debug_misc( __FUNCTION__, "multiple contacts found!\n" );
  }

  return blist;
}

/*
 * build an xml fragment representing a contact
 */
static xmlNodePtr build_contact( gchar *host, gchar *nick, gchar *desc, gchar *principal ) {
  xmlNodePtr contact, node, content;
  gchar *url;

  if ( principal == NULL ) {
    url = g_strdup_printf( "http://%s/instmsg/aliases/%s", host, nick );
  } else {
    url = g_strdup( principal );
  }
  contact = xmlNewNode( NULL, BAD_CAST "r:contact" );
  node = xmlNewNode( NULL, BAD_CAST "d:href" );
  content = xmlNewText( BAD_CAST url );
  xmlAddChild( node, content );
  xmlAddChild( contact, node );

  node = xmlNewNode( NULL, BAD_CAST "r:description" );
  if ( desc ) {
    content = xmlNewText( BAD_CAST desc );
    xmlAddChild( node, content );
  }
  xmlAddChild( contact, node );

  return contact;
}

/*
 * look up a buddy by rvp name, principal, or session-id
 */
static GaimBuddy *rvp_find_buddy( GaimConnection *gc, gchar *name,
                                  gchar *principal, gchar *session ) {
  GaimBuddyList *blist;
  GaimBlistNode *group, *cnode, *bnode;
  GaimBuddy *retval = NULL;

  blist = gaim_get_blist();
  for ( group = blist->root; group; group = group->next ) {
    for ( cnode = group->child; cnode; cnode = cnode->next ) {
      if ( !GAIM_BLIST_NODE_IS_CONTACT(cnode)) {
        continue;
      }
      for ( bnode = cnode->child; bnode; bnode = bnode->next ) {
        RVPBuddy *target;

        if (!GAIM_BLIST_NODE_IS_BUDDY( bnode )) {
          continue;
        }
        retval = (GaimBuddy *)bnode;

        if ( retval->account != gc->account ) {
          retval = NULL;
          continue;
        }

        if ( name != NULL ) {
          if ( !strcmp( retval->name, name )) {
            goto gotcha;
          } else {
            continue;
          }
        }

        /*
         * Any thing below here requires that the RVPBuddy structure
         * is valid.
         */
        target = retval->proto_data;

        /*
         * async notifications can cause this to be triggered
         * before we've set up the proto_data for the buddies
         */
        if ( target == NULL ) {
          retval = NULL;
          continue;
        }

        if ( principal != NULL ) {
          if ( target->principal != NULL &&
               !strcmp( target->principal, principal )) {
            goto gotcha;
          }
        }

        if ( session != NULL ) {
          if ( target->sessionid != NULL &&
               !strcmp( target->sessionid, session )) {
            goto gotcha;
          }
        }

        retval = NULL;
      }
    }
  }

 gotcha:
    return retval;
}

/*
 * parse the head of an XML response. ROOT is the expected root node.
 */
static xmlDocPtr parse_preamble( const gchar *buffer, gint size, const gchar *root, xmlNodePtr *cur ) {
  xmlDocPtr doc = NULL;

#ifdef LOUD
  gaim_debug_misc( __FUNCTION__, "\n%*.*s\n", size, size, buffer );
#endif

  /* by preference, use xmlReadMemory as it can be told not to whine
     about the invalid 'DAV:' namespace */
#ifdef HAVE_XMLREADMEMORY
  if (( doc = xmlReadMemory( buffer, size, NULL, NULL, XML_PARSE_NOWARNING )) == NULL ) {
    gaim_debug_error( __FUNCTION__, "Document not parsed successfully.\n" );
    return NULL;
  }
#else
  if (( doc = xmlParseMemory( buffer, size )) == NULL ) {
    gaim_debug_error( __FUNCTION__, "Document not parsed successfully.\n" );
    return NULL;
  }
#endif

  if (( *cur = xmlDocGetRootElement( doc )) == NULL ) {
    gaim_debug_error( __FUNCTION__, "empty document\n" );
    xmlFreeDoc( doc );
    return NULL;
  }

  if ( xmlStrcmp( (*cur)->name, BAD_CAST root )) {
    gaim_debug_error( __FUNCTION__, "unexpected doc root\n" );
    xmlFreeDoc( doc );
    return NULL;
  }

  return doc;
}

/*
 * Parse a SUBSCRIPTIONS response to get our buddy list
 */
static void parse_subscriptions( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  xmlDocPtr doc;
  xmlNodePtr cur, subscription;

  gaim_debug_misc( __FUNCTION__, "enter\n" );

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len,
                              "subscriptions", &cur )) == NULL ) {
    goto out;
  }

  if (( subscription = cur->xmlChildrenNode ) == NULL ) {
    gaim_debug_misc( __FUNCTION__, "no subscriptions\n" );
    goto out;
  }

  /*
  ELEMENT r:subscription
    ELEMENT r:subscription-id
      TEXT
        content=1765
    ELEMENT a:principal
      ELEMENT a:rvp-principal
        TEXT
          content=http://host/instmsg/aliases...
    ELEMENT d:href
      TEXT
        content=http://host/instmsg/aliases...
    ELEMENT d:timeout
      TEXT
        content=13135
  */

  while( subscription ) {
    xmlNodePtr ptr = subscription->xmlChildrenNode;
    gint subs_id, timeout;
    gchar *principal = NULL, *href = NULL;

    while( ptr ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "subscription-id" )) {
        xmlChar *sid = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        subs_id = atoi((gchar *)sid );
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "href" )) {
        href = (gchar *)xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "timeout" )) {
        xmlChar *to = xmlNodeListGetString( doc, ptr->xmlChildrenNode, 1 );
        timeout = atoi((gchar *)to );
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "principal" )) {
        xmlNodePtr rvpp = ptr->xmlChildrenNode;
        if ( rvpp != NULL && !xmlStrcmp( rvpp->name, BAD_CAST "rvp-principal" )) {
          principal = (gchar *)xmlNodeListGetString( doc, rvpp->xmlChildrenNode, 1 );
        }
      } else {
        /* unknown item, ignore for now FIXME */
        gaim_debug_error( __FUNCTION__, "what's a %s?\n", (gchar *)ptr->name );
      }

      ptr = ptr->next;
    }

    /* this is the only thing we really need to retrieve the
       subscription */
    if ( principal != NULL ) {
      gchar **split = NULL;
      gint elt = 0;
      GaimBuddy *buddy;
      gboolean temp = FALSE;

      /* hmm. this doesn't currently recognise ME... */
      if (( buddy = rvp_find_buddy( gc, NULL, principal, NULL )) == NULL ) {
        buddy = g_new0( GaimBuddy, 1 );
        temp = TRUE;
      }

      if ( buddy->proto_data == NULL ) {
        buddy->proto_data = g_new0( RVPBuddy, 1 );
        ((RVPBuddy *)buddy->proto_data)->buddy = buddy;
      }

      ((RVPBuddy *)buddy->proto_data)->principal = g_strdup( principal );

      /* funnily enough, doing this causes all manner of
         trouble. don't do it.
      ((RVPBuddy *)buddy->proto_data)->subs_id = subs_id;
      */

      if ( temp ) {
        buddy = rvp_get_temp_buddy_create( gc, buddy );
      }

      /* From Microsoft's RVP doc:

      There are two types of URLs: logical URLs and physical URLs. A
      logical URL is the default and determines which domain the node
      is located on. For example, an e-mail address such as
      beverlyj@domain1.com could map to a logical URL
      http://im.domain1.com/instmsg/aliases/beverlyj.

      A physical URL can be the same as a logical URL, or it can
      provide extra information as to which home server is responsible
      for the entity. A physical URL is used when a router redirects
      requests for a specific entity. For example, if the e-mail
      address garrettv@domain2.com is within a domain with a router
      and several home servers, any requests to the router can be
      redirected to the physical URL
      http://imhome1.domain2.com/instmsg/local/im.domain2.com/instmsg/aliases/garrettv.
      This physical URL is then used to access the node at the home
      server imhome1.domain2.com.
      */
      split = g_strsplit( principal, "/", 0 );
      while( split[elt] != NULL ) { elt++; }
      if ( strstr( principal, "/instmsg/local" ) != NULL ) {
        /* physical URL */
        if ( elt == 9 ) {
          buddy->name = g_strconcat( split[8], "@", split[2], NULL );
        } else {
          gaim_debug_error( __FUNCTION__, "can't parse principal %s\n",
                            principal );
        }
      } else {
        if ( elt == 6 ) {
          buddy->name = g_strconcat( split[5], "@", split[2], NULL );
        } else {
          gaim_debug_error( __FUNCTION__, "can't parse principal %s\n",
                            principal );
        }
      }

      if ( split ) {
        g_strfreev( split );
      }
    } else {
      gaim_debug_error( __FUNCTION__, "principal null\n" );
    }

    subscription = subscription->next;
  }

 out:
  if ( doc != NULL ) {
    xmlFree( doc );
  }

  gaim_debug_misc( __FUNCTION__, "out\n" );
}

/*
 * Parse an ACL collection
 */
static void parse_acls( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  RVPData *rd = gc->proto_data;
  xmlDocPtr doc;
  xmlNodePtr cur, acl, ace;

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len, "rvpacl",
                              &cur )) == NULL ) {
    return;
  }

  if (( acl = cur->xmlChildrenNode ) == NULL ) {
    gaim_debug_error( __FUNCTION__, "no acls\n" );
    goto out;
  }

  while( acl ) {
    ace = acl->xmlChildrenNode;

    while( ace ) {
      GaimBuddy *buddy = NULL;
      xmlChar *rvp_principal = NULL;
      xmlNodePtr credentials = NULL;
      xmlNodePtr bits = ace->xmlChildrenNode;

      /* skip the 'inheritance: none' crap */
      if ( xmlStrcmp( ace->name, BAD_CAST "ace" )) {
        gaim_debug_misc( __FUNCTION__, "skipping %s\n", (gchar *)ace->name );
        goto loop;
      }

      /* the horror, the horror */
      if ( !xmlStrcmp( bits->name, BAD_CAST "principal" )) {
        xmlNodePtr pbits = bits->xmlChildrenNode;
        while( pbits != NULL ) {
          if ( !xmlStrcmp( pbits->name, BAD_CAST "rvp-principal" )) {
            if (( rvp_principal =
                  xmlNodeListGetString( doc, pbits->xmlChildrenNode, 1 ))
                != NULL ) {
              if ( !xmlStrcmp( BAD_CAST rd->me.principal, rvp_principal )) {
                buddy = rd->me.buddy;
              } else {
                buddy = rvp_find_buddy( gc, NULL, (gchar *)rvp_principal,
                                        NULL );

                if ( buddy == NULL ) {
                  /* need to fake enough of a GaimBuddy structure to
                     keep this happy, which maybe I should do elsewhere... */
                  GaimBuddy *fake = g_new0( GaimBuddy, 1 );
                  RVPBuddy *faker = g_new0( RVPBuddy, 1 );
                  fake->proto_data = faker;
                  faker->principal = (gchar *)rvp_principal;
                  buddy = rvp_get_temp_buddy_create( gc, fake );
                }
              }
            }
          } else if ( !xmlStrcmp( pbits->name, BAD_CAST "credentials" )) {
            credentials = pbits->xmlChildrenNode;
          }
          pbits = pbits->next;
        }
      }

      if ( buddy ) {
        RVPBuddy *rbuddy = buddy->proto_data;
        xmlNodePtr grants = NULL, denies = NULL;

        while( bits ) {
          if ( !xmlStrcmp( bits->name, BAD_CAST "grant" )) {
            grants = bits->xmlChildrenNode;
            gaim_debug_misc( __FUNCTION__, "got grant\n" );
          } else if ( !xmlStrcmp( bits->name, BAD_CAST "deny" )) {
            denies = bits->xmlChildrenNode;
            gaim_debug_misc( __FUNCTION__, "got deny\n" );
          }
          bits = bits->next;
        }

        while ( credentials ) {
          if ( !xmlStrcmp( credentials->name, BAD_CAST "assertion" )) {
            rbuddy->acl.credentials.assertion = TRUE;
          } else if ( !xmlStrcmp( credentials->name, BAD_CAST "digest" )) {
            rbuddy->acl.credentials.digest = TRUE;
          } else if ( !xmlStrcmp( credentials->name, BAD_CAST "ntlm" )) {
            rbuddy->acl.credentials.ntlm = TRUE;
          }

          credentials = credentials->next;
        }

        while ( grants ) {
          if ( !xmlStrcmp( grants->name, BAD_CAST "list" )) {
            rbuddy->acl.list = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "read" )) {
            rbuddy->acl.read = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "write" )) {
            rbuddy->acl.write = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "send-to" )) {
            rbuddy->acl.send_to = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "receive-from" )) {
            rbuddy->acl.receive_from = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "readacl" )) {
            rbuddy->acl.readacl = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "writeacl" )) {
            rbuddy->acl.writeacl = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "presence" )) {
            rbuddy->acl.presence = TRUE;
          } else if ( !xmlStrcmp( grants->name, BAD_CAST "subscriptions" )) {
            rbuddy->acl.subscriptions = TRUE;
          } else if ( !xmlStrcmp( grants->name,
                                  BAD_CAST "subscribe-others" )) {
            rbuddy->acl.subscribe_others = TRUE;
          } else {
            gaim_debug_warning( __FUNCTION__, "unrecognised ACL %s\n",
                                grants->name );
          }

          grants = grants->next;
        }

        while ( denies ) {
          if ( !xmlStrcmp( denies->name, BAD_CAST "list" )) {
            rbuddy->acl.list = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "read" )) {
            rbuddy->acl.read = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "write" )) {
            rbuddy->acl.write = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "send-to" )) {
            rbuddy->acl.send_to = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "receive-from" )) {
            rbuddy->acl.receive_from = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "readacl" )) {
            rbuddy->acl.readacl = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "writeacl" )) {
            rbuddy->acl.writeacl = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "presence" )) {
            rbuddy->acl.presence = FALSE;
          } else if ( !xmlStrcmp( denies->name, BAD_CAST "subscriptions" )) {
            rbuddy->acl.subscriptions = FALSE;
          } else if ( !xmlStrcmp( denies->name,
                                  BAD_CAST "subscribe-others" )) {
            rbuddy->acl.subscribe_others = FALSE;
          } else {
            gaim_debug_warning( __FUNCTION__, "unrecognised ACL %s\n",
                                denies->name );
          }

          denies = denies->next;
        }

        if ( buddy == NULL ) {
          rvp_dump_buddy( rd->me.buddy );
        } else {
          rvp_dump_buddy( buddy );
        }
      } else {
        if ( rvp_principal != NULL ) {
          /* shouldn't ever see this as we're faking up a buddy struct
             above */
          gaim_debug_misc( __FUNCTION__, "can't find buddy for %s\n",
                           (gchar *)rvp_principal );
          xmlDebugDumpNode( stderr, ace, TRUE );
        } else {
          /* HMM. Maybe I should create a fake buddy for this as it'd
             give me plenty space to keep the ACLs */
          gaim_debug_misc( __FUNCTION__,
                           "Null Principal, probably generic acl\n" );
          /* xmlDebugDumpNode( stderr, ace, TRUE ); */
        }
      }

    loop:
      ace = ace->next;
    }

    acl = acl->next;
  }

 out:
  xmlFreeDoc( doc );
}

/*
 * Parse a 207 Multistatus response and apply its data to whatever
 * we've got in core.
 */
static void parse_multistatus( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  xmlDocPtr doc;
  xmlNodePtr cur, response;

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len,
                              "multistatus", &cur )) == NULL ) {
    return;
  }

  /* children of a multistatus:
   * - response {set}
   *   - href of principal concerned
   *   - propstat
   *     - prop
   *       - r:state, d:displayname
   *     - status
   */
  response = cur->xmlChildrenNode;

  while ( response != NULL ) {
    xmlNodePtr ptr, href = NULL, propstat = NULL, prop = NULL, status = NULL;
    xmlChar *val;
    RVPData *rvpdata = gc->proto_data;
    RVPBuddy *target = NULL, *metoo = NULL;
    GaimBuddy *buddy = NULL;

    if ( xmlStrcmp( response->name, BAD_CAST "response" )) {
      gaim_debug_error( __FUNCTION__, "expected <response> here\n" );
      goto loop;
    }

    ptr = response->xmlChildrenNode;
    while ( ptr != NULL ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "href" )) {
        href = ptr;
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "propstat" )) {
        propstat = ptr;
      } else {
        gaim_debug_error( __FUNCTION__, "unknown response part\n" );
      }

      ptr = ptr->next;
    }

    if ( href == NULL || propstat == NULL ) {
      goto loop;
    }

    if (( val =
          xmlNodeListGetString( doc, href->xmlChildrenNode, 1 )) == NULL ) {
      goto loop;
    }

    /*
     * using metoo here in case you've got yourself on your buddy
     * list. MSM won't allow you to do that, but Gaim will.
     */
    target = NULL;
    metoo = NULL;

    if ( !strcmp( rvpdata->me.principal, (char *)val )) {
      metoo = &(rvpdata->me);
    }

    buddy = rvp_find_buddy( gc, NULL, (char *)val, NULL );
    if ( buddy == NULL ) {
      target = metoo;
    } else {
      target = buddy->proto_data;
    }

    if ( target == NULL ) {
      gaim_debug_warning( __FUNCTION__, "don't know this principal (%s)\n", (char *)val );
      goto loop;
    } else if ( target == metoo ) {
      gaim_debug_misc( __FUNCTION__, "setting state for myself\n" );
    }

    /* now handle the props */

    /* first, get the status */
    ptr = propstat->xmlChildrenNode;
    while( ptr != NULL ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "prop" )) {
      } else if ( !xmlStrcmp( ptr->name, BAD_CAST "status" )) {
        status = ptr;
      } else {
        gaim_debug_error( __FUNCTION__, "unknown response part\n" );
      }

      ptr = ptr->next;
    }

    if ( status == NULL ) {
      gaim_debug_error( __FUNCTION__, "No status in property\n" );
      goto loop;
    }

    /* xxx extract status code */

    /* now do the props */
    ptr = propstat->xmlChildrenNode;
    while( ptr != NULL ) {
      if ( !xmlStrcmp( ptr->name, BAD_CAST "prop" )) {
        prop = ptr->xmlChildrenNode;

        while( prop != NULL ) {
          if ( !xmlStrcmp( prop->name, BAD_CAST "displayname" )) {
            if (( val =
                  xmlNodeListGetString( doc, prop->xmlChildrenNode, 1 ))
                != NULL ) {
              if ( target->buddy->server_alias != NULL ) {
                g_free( target->buddy->server_alias );
              }
              target->buddy->server_alias = g_strdup( (gchar *)val );

              if ( metoo != NULL && target != metoo ) {
                if ( metoo->buddy->server_alias != NULL ) {
                  g_free( metoo->buddy->server_alias );
                }
                metoo->buddy->server_alias = g_strdup( (gchar *)val );
                gaim_account_set_alias( gaim_connection_get_account( gc ),
                                        g_strdup(( gchar *) val ));
              }

              xmlFree( val );
            }

            /* update the buddy list view */
            if ( target != metoo && target->buddy->server_alias != NULL ) {
              gaim_blist_update_buddy_icon( target->buddy );
            }

          } else if ( !xmlStrcmp( prop->name, BAD_CAST "state" )) {
            /* big state or little state? */
            xmlNodePtr state = NULL, snode = prop->xmlChildrenNode;

            while( snode ) {
              if ( !xmlStrcmp( snode->name, BAD_CAST "leased-value" )) {
                /* big state */
                xmlNodePtr bits = snode->xmlChildrenNode;

                while( bits ) {
                  if ( !xmlStrcmp( bits->name, BAD_CAST "value" )) {
                    state = bits->xmlChildrenNode;
                  } else if ( !xmlStrcmp( bits->name, BAD_CAST "timeout" )) {
                    xmlChar *to = xmlNodeListGetString( doc, bits->xmlChildrenNode , 1 );
                    if ( to != NULL ) {
                      rvpdata->view_expiry = atol( (gchar *)to ) + time( NULL );
                      gaim_debug_misc( __FUNCTION__, "view_id expires in %u seconds (%s)\n", atol((gchar *)to ), to );
                    }
                  }

                  bits = bits->next;
                }
              } else if ( !xmlStrcmp( snode->name, BAD_CAST "view-id" )) {
                gchar *viewid = (gchar *)xmlNodeListGetString( doc, snode->xmlChildrenNode, 1 );
                if ( viewid != NULL ) {
                  rvpdata->view_id = atol( viewid );
                } else {
                  gaim_debug_error( __FUNCTION__, "can't parse view-id\n" );
                }
              } else {
                /* little state */
                state = snode;
              }

              snode = snode->next;
            }

            if ( state != NULL && target != metoo ) {
              rvp_set_buddy_status( gc, target->buddy->name,
                                    (char *)state->name );
            }
          } else {
            xmlChar *v = xmlNodeListGetString( doc, prop->xmlChildrenNode, 1 );

            if ( !xmlStrcmp( prop->name, BAD_CAST "email" )) {
              target->email = g_strdup( (gchar *)v );
            } else if ( !xmlStrcmp( prop->name, BAD_CAST "mobile-state" )) {
              target->mobile_state = atoi((gchar *)v );
            } else if ( !xmlStrcmp( prop->name, BAD_CAST "mobile-description" )) {
              target->mobile_description = g_strdup((gchar *)v);
            } else {
              gaim_debug_warning( __FUNCTION__, "unknown property %s = %s\n",
                                  (gchar *)prop->name, (gchar *)v );
            }
          }
          prop = prop->next;
        }
      }

      ptr = ptr->next;
    }

  loop:
    response = response->next;
  }

  xmlFreeDoc( doc );
}

/*
 * extract a given header from the header block and return its content
 *
 * This is a bit wasteful in that it reparses the headers every time
 * it's called, but hey. Also, it expects both inputs to be null-terminated.
 */
static gchar *get_header_content( gchar *hblock, gchar *header ) {
  gchar **headers = NULL;
  gchar *retval = NULL, *headercolon = NULL;

  headercolon = g_strconcat( header, ":", NULL );

  /* fast return */
  if ( !strstr( hblock, headercolon )) {
    g_free( headercolon );
    return NULL;
  }

  /* technically \r\n, but be lenient in what you accept &c. */
  headers = g_strsplit( hblock, "\n", 0 );

  if ( headers != NULL ) {
    gint i;
    for ( i = 0; headers[ i ]; i++ ) {
      /* fixme hardcoded == teh wrongness */
      /* the correct way to handle this is to either fold the headers
         together, or to allow for a param specifying which instance
         of a header I'm interested in */
      if ( !strncmp( headers[i], "WWW-Authenticate: Negotiate",
                     strlen( "WWW-Authenticate: Negotiate" ))) {
        continue;
      }

      if ( !strncmp( headers[ i ], headercolon, strlen( headercolon ))) {
        retval = g_strdup( &headers[i][strlen( headercolon )]);
        /* xxx glib doc doesn't make it clear if this is done in-place */
        retval = g_strstrip( retval );

        break;
      }
    }

    g_strfreev( headers );
  }
  g_free( headercolon );

  return retval;
}

/*
 * Build a minimal buddy structure from an email address
 */
static GaimBuddy *rvp_buddy_from_address( GaimConnection *gc, gchar *fullname ) {
  GaimBuddy *retval = NULL;
  RVPBuddy *rb = NULL;
  gchar **bits = g_strsplit( fullname, "@", 2 );
  gchar *srvname, *host = NULL;
  srvrec *service = NULL;
  gint port = RVP_PORT;

  if ( bits == NULL || bits[0] == NULL ) {
    goto out;
  }

  if ( bits[1] == NULL ) {
    const gchar *newname = rvp_normalize( gaim_connection_get_account( gc ),
                                    fullname );
    g_strfreev( bits );
    bits = g_strsplit( newname, "@", 2 );
  }

  host = g_strdup( bits[1] );

  srvname = g_strconcat( "_rvp._tcp.", bits[1], NULL );
  service = gethostbysrv( srvname, NULL );
  g_free( srvname );

  if ( service->host == NULL ) {
    struct hostent *h;
    if (( h = gethostbyname( bits[1] )) == NULL ) {
      goto out;
    }
  } else {
    g_free( host );
    host = g_strdup( service->host );
    port = service->port;
  }

  retval = g_new0( GaimBuddy, 1 );
  retval->proto_data = g_new0( RVPBuddy, 1 );
  rb = retval->proto_data;
  if ( port == RVP_PORT ) {
    rb->principal = g_strdup_printf( "http://%s/instmsg/aliases/%s",
                                     host, bits[0] );
  } else {
    rb->principal = g_strdup_printf( "http://%s:%d/instmsg/aliases/%s",
                                     host, port, bits[0] );
  }
  retval->name = g_strconcat( bits[0], "@", host, NULL );

 out:
  if ( host ) {
    g_free( host );
  }
  if ( bits != NULL ) {
    g_strfreev( bits );
  }
  return retval;
}

/*
 * parse a notification
 */
static gint parse_notify( GaimConnection *gc, GaimFetchUrlData *gfud ) {
  xmlDocPtr doc = NULL;
  xmlNodePtr cur, notification;
  gint retval = 200;
  GaimBuddy *buddy = NULL, *from_contact = NULL, *to_contact = NULL,
    **to_contacts = NULL;
  RVPData *rd = gc->proto_data;

  if (( doc = parse_preamble( gfud->response.webdata, gfud->len,
                              "notification", &cur )) == NULL ) {
    retval = 400;
    goto out;
  }

  notification = cur->xmlChildrenNode;

  while( notification != NULL ) {
    /* what type of notification? */
    if ( !xmlStrcmp( notification->name, BAD_CAST "message" ) ||
         !xmlStrcmp( notification->name, BAD_CAST "propnotification" )) {
      xmlNodePtr ptr = notification->xmlChildrenNode;
      xmlNodePtr from = NULL, to = NULL, msg = NULL, props = NULL;
      xmlChar *val = NULL;

      gaim_debug_misc( __FUNCTION__, "Notify type: %s\n",
                       (gchar *)notification->name );

      while( ptr ) {
        if ( !xmlStrcmp( ptr->name, BAD_CAST "notification-from" )) {
          from = ptr->xmlChildrenNode;
        } else if ( !xmlStrcmp( ptr->name,
                                BAD_CAST "notification-to" )) {
          to = ptr->xmlChildrenNode;
        } else if ( !xmlStrcmp( ptr->name, BAD_CAST "msgbody" )) {
          msg = ptr->xmlChildrenNode;
        } else if ( !xmlStrcmp( ptr->name, BAD_CAST "propertyupdate" )) {
          props = ptr->xmlChildrenNode;
        } else {
          gaim_debug_error( __FUNCTION__, "unknown notification part\n" );
        }
        ptr = ptr->next;
      }

      /* nothing to do? bail! */
      if ( !msg && !props ) {
        goto loop;
      }

      if ( !from ) { /* go digging for a session id! */
        goto loop;
      } else {
        /* who the hell are you? */
        GaimBuddy **from_contacts = parse_contact( doc, from );
        if ( from_contacts == NULL ) {
          /* xxx */
          gaim_debug_error( __FUNCTION__, "no/unparseable sender\n" );
          goto loop;
        } else {
          /* we do not expect to see multiple senders... */
          from_contact = from_contacts[ 0 ];
          buddy = rvp_find_buddy( gc, NULL,
                                  ((RVPBuddy *)from_contact->proto_data)->principal, NULL );
        }
      }

      if ( to ) {
        to_contacts = parse_contact( doc, to );
        if ( to_contacts == NULL ) {
          gaim_debug_error( __FUNCTION__, "no/unparseable recipient\n" );
        } else {
          to_contact = to_contacts[ 0 ];
        }
      }

      if ( msg != NULL && !xmlStrcmp( msg->name, BAD_CAST "mime-data" )) {
        if (( val =
              xmlNodeListGetString( doc, msg->xmlChildrenNode, 1 )) != NULL ) {
          gchar **parts = NULL;

          if ( strstr( (gchar *)val, "\r\n\r\n" )) {
            parts = g_strsplit( (gchar *)val, "\r\n\r\n", 2 );
          } else if ( strstr( (gchar *)val, "\n\n" )) {
            parts = g_strsplit( (gchar *)val, "\n\n", 2 );
          } else {
            /* what sort of mutant are you, exactly? */
            gaim_debug_error( __FUNCTION__,
                              "can't find header delimiter in \n%s\n",
                              (gchar *)val );
          }

          if ( parts[0] != NULL ) {
            gchar *typist = NULL;
            gchar *sessid = NULL;
            gint id = 0; /* for chats */
            gboolean is_chat = FALSE;
            gint msgtype = RVP_MSG_UNKNOWN;
            gchar *contenttype;

            contenttype = get_header_content( parts[0], "Content-Type" );
            if ( contenttype ) {
              if ( !strncmp( contenttype, "text/plain", strlen( "text/plain" ))) {
                /* the rest may or may not specify a charset */
                /* fixme should grab it if it's there and behave
                   accordingly */
                msgtype = RVP_MSG_IM;
              } else if ( !strncmp( contenttype, "text/x-msmsgscontrol",
                                    strlen( "text/x-msmsgscontrol" ))) {
                msgtype = RVP_MSG_TYPING;
              } else if ( !strncmp( contenttype, "text/x-msmsgsinvite",
                                    strlen( "text/x-msmsgsinvite" ))) {
                /* for now... */
                msgtype = RVP_MSG_IM;
              } else if ( !strncmp( contenttype, "text/x-imleave",
                                    strlen( "text/x-imleave" ))) {
                msgtype = RVP_MSG_LEAVE;
              }
            }

            sessid = get_header_content( parts[0], "Session-Id" );
            typist = get_header_content( parts[0], "TypingUser" );

            /* todo:
               MIME-Version (ignored)
               X-MMS-IM-Format */

            if ( sessid != NULL ) {
              /* get rid of the curlies */
              if ( sessid[0] == '{' ) {
                sessid = memmove( sessid, &sessid[1], strlen( sessid ) - 1 );
              }
              if ( sessid[strlen( sessid ) - 2] == '}' ) {
                sessid[strlen( sessid ) - 2] = '\0';
              }

              gaim_debug_misc( __FUNCTION__, "session id %s\n", sessid );

              /* see if this is a multi-user chat! */
              if ( rd->chats ) {
                GaimConversation *conv = g_hash_table_lookup( rd->chats, sessid );
                if ( conv != NULL ) {
                  if ( gaim_conversation_get_type( conv ) == GAIM_CONV_CHAT ) {
                    id = gaim_conv_chat_get_id( GAIM_CONV_CHAT( conv ));
                    is_chat = TRUE;
                  } else {

                  }
                }
              }

              /* you've been press-ganged into a multi-user chat */
#if 0
              if ( !is_chat && to_contacts[0] != NULL &&
                   to_contacts[1] != NULL ) {
                GHashTable *chat = g_hash_table_new( g_str_hash, g_str_equal );
                is_chat = TRUE;

                g_hash_table_insert( chat, g_strdup( "sessid" ), sessid );
                g_hash_table_insert( chat, g_strdup( "from" ), from_contact );
                g_hash_table_insert( chat, g_strdup( "other" ), to_contacts );

                serv_join_chat( gc, chat );
              }
#endif

              /* see if we've got a handle to work with */
              if ( from_contact ) {
                RVPBuddy *p;
                if ( buddy == NULL ) {
                  buddy = rvp_get_temp_buddy_create( gc, from_contact );
                }

                p = buddy->proto_data;

                if ( p ) {
                  /* trash the old session ID */
                  if ( p->sessionid != NULL ) {
                    g_free( p->sessionid );
                  }
                  p->sessionid = g_strdup( sessid );
                } else {
                  /* can't happen, ish */
                  gaim_debug_error( __FUNCTION__, "no proto data\n" );
                }
              } else {
                /* errrr. */
              }
            }

            if ( parts[1] != NULL ) {
              if ( msgtype == RVP_MSG_IM ) {
                if ( is_chat ) {
                  serv_got_chat_in( gc, id, buddy->name, 0, (char *)parts[1],
                                    time( NULL ));
                } else {
                  serv_got_im( gc, buddy->name, (char *)parts[1], 0,
                               time( NULL ));
                  /*
                  GaimConversation *conv;
                  conv = gaim_find_conversation_with_account( buddy->name, gaim_connection_get_account( gc ));
                  g_hash_table_replace( rd->chats, sessid, conv );
                  */
                }
                /* fixme auto reply if away */
              } else if ( msgtype == RVP_MSG_TYPING ) {
                /* hrm. not only is there an at-sign in the typist, but
                   the domain is technically incorrect. I could probably
                   look it up in the buddy list or via the session ID,
                   but for now: */
                if ( typist == NULL ) {
                  /* this happens when a remote user pops up a dialog
                     box to talk to you but hasn't started typing
                     yet. this can be used to pop a dialog box
                     asking if you want to receive an unsolicited
                     message. if there are multiple recipients for
                     this then it's an invite to a multi-user chat */
                  if ( to_contacts[ 1 ] != NULL ) {
                    GHashTable *chat = g_hash_table_new( g_str_hash, g_str_equal );

                    g_hash_table_insert( chat, "sessid", sessid );
                    g_hash_table_insert( chat, "from", from_contact );
                    g_hash_table_insert( chat, "other", to_contacts );

                    /* the proper windows-like logic here is to only
                       do a serv_join_chat here if we're already in
                       conversation; otherwise we should note the
                       sessid and punt until we receive a message from
                       this conversation */
                    /* serv_got_chat_invite( gc, "Multi-User Conversation",
                       from_contact->name, NULL, chat ); */
                    serv_join_chat( gc, chat );
                  } else {
                    /* nothing to do for now, since ACLs seem to be
                       managed on the server? */
                  }
                } else {
                  if ( buddy != NULL && buddy->name == NULL ) {
                    gaim_debug_misc( __FUNCTION__, "added buddy name %s\n",
                                     typist );
                    buddy->name = g_strdup( typist );
                  }

                  serv_got_typing( gc, typist, EXCHANGE_TYPING_TIMEOUT,
                                   GAIM_TYPING );
                }
              } else if ( msgtype == RVP_MSG_LEAVE ) {
                /* user has left a chat */
                rvp_chat_user_left( gc, id, from_contact );
              } else {
                gaim_debug_misc( __FUNCTION__, "unknown messagetype %s\n",
                                 contenttype );
              }
            } else {
              gaim_debug_misc( __FUNCTION__, "empty message\n" );
            }

            g_strfreev( parts );
          } else {
            /* no parts, but we've already warned */
          }
          xmlFree( val );
        }
      } else {
        xmlNodePtr prop = props->xmlChildrenNode;
        while( prop ) {
          xmlNodePtr name = prop->xmlChildrenNode;
          xmlNodePtr value = NULL;

          if ( name != NULL ) {
            value = name->xmlChildrenNode;
          } else {
            /* we expect a null here, as it happens */
            val = xmlNodeListGetString( doc, name, 1 );
            if ( val != NULL ) {
              gaim_debug_misc( __FUNCTION__, "non-null prop\n" );
            }
            goto proploop;
          }

          /* we only care about buddies we can recognise from our list
             or our conversation windows */
          if ( buddy == NULL ) {
            buddy = rvp_get_temp_buddy_create( gc, from_contact );
          }

          if ( !xmlStrcmp( name->name, BAD_CAST "state")) {
            if ( buddy != NULL ) {
              rvp_set_buddy_status( gc, buddy->name, (gchar *)value->name );
            } else {
              gaim_debug_misc( __FUNCTION__,
                               "got state for unknown buddy %s\n",
                               ((RVPBuddy *)from_contact->proto_data)->principal );
            }
          } else {
            gaim_debug_misc( __FUNCTION__, "unknown property %s\n",
                             name->name );
          }

        proploop:
          prop = prop->next;
        }
      }
    } else {
      fprintf( stderr, "unhandled notification type " );
      xmlDebugDumpString( stderr, notification->name );
      fprintf( stderr, "\n" );
      xmlDebugDumpNode( stderr, notification, 0 );
    }

  loop:
    notification = notification->next;
  }

 out:
  if ( doc ) {
    xmlFreeDoc( doc );
  }
  gaim_debug_misc( __FUNCTION__, "exit %d\n", retval );

  return retval;
}

/*
 * Using ACCOUNT, generate a normalised version of the username S
 */
static const char *rvp_normalize( const GaimAccount *account, const char *s ) {
  static char buf[BUF_LEN]; /* gaim requirement, alas */
  GaimBuddy *b;
  gchar **bits;
  gchar *host = NULL;

  g_return_val_if_fail( s != NULL, NULL );

  gaim_debug_misc( __FUNCTION__, "normalizing %s\n", s );

  bits = g_strsplit( s, "@", 2 );

  if ( bits == NULL || bits[0] == NULL ) {
    gaim_debug_error( __FUNCTION__, "failed to split %s\n", s );
    abort();
  }

  if ( bits[1] == NULL ) {
    /* no hostname specified: use the same one as our account */
    GaimConnection *gc = gaim_account_get_connection( account );
    RVPData *rd;
    if ( gc != NULL ) {
      rd = gc->proto_data;
      if ( rd->host != NULL ) {
        host = g_strdup( rd->host );
      } else {
        gaim_debug_error( __FUNCTION__, "doooh 1\n" );
        abort();
      }
    } else {
      /* GaimConnection is unavailable, so we'll need to use
         GaimAccount */
      gchar **bits2 = g_strsplit( gaim_account_get_username( account ),
                                  "@", 2 );
      if ( bits2 == NULL || bits[0] == NULL ) {
        gaim_debug_error( __FUNCTION__, "failed to splioot %s\n",
                          gaim_account_get_username( account ));
        abort();
      }
      if ( bits2[1] == NULL ) {
        /* you're really starting to annoy me */
        host = g_strdup( gaim_account_get_string( account, "host", NULL ));
      } else {
        host = g_strdup( bits2[1] );
      }
      g_strfreev( bits2 );
    }
  } else {
    host = g_strdup( bits[1] );
  }

  g_snprintf( buf, sizeof( buf ), "%s@%s", bits[0], host );
  g_strfreev( bits );
  g_free( host );

  /* now expand the host out */
  b = rvp_buddy_from_address( gaim_account_get_connection( account ), buf );
  if ( b != NULL ) {
    memcpy( buf, b->name, strlen( b->name ));
  }
  rvp_free_buddy( b );

  gaim_debug_misc( __FUNCTION__, "result is %s\n", buf );

  return buf;
}

/*
 * generate a session ID
 * from the expired RFC:
 *
 * 10.2.19.  Session-ID
 *
 *  Sent with a NOTIFY message when there is no subscription-ID; used
 *  to maintain context in replies to that notification.
 *
 * The Microsoft documentation has no information on this. From
 * observation, Microsoft Messenger preserves session ID only for as
 * long as its login persists. As soon as you sign out and back in
 * again, it creates a new session ID.
 */
static gchar *rvp_get_sessid( void ) {
  gchar *my_sessid = g_malloc0( 37 );
  int i, r;
  int st = 1;

  init_seed( st );
  for (i = 0; i < 36; i++) {
    if ( i == 8 || i == 13 || i == 18 || i == 23 )
      my_sessid[i] = '-';
    else {
      r = random_integer( 0, 15 );
      my_sessid[ i ] = HEX_ELM[ r ];
    }
  }

  gaim_debug_misc( __FUNCTION__, "generated new ID %s\n", my_sessid );

  return my_sessid;
}

/*
 * Return the name of an icon for this protocol
 * not really sure why this is a /function/...
 */
static const char *rvp_list_icon( GaimAccount *a, GaimBuddy *b ) {
  return "msn"; /* fixme do an rvp icon ? */
}

/*
 * Figure out what state a buddy is in
 */
static void rvp_list_emblems( GaimBuddy *b, char **se, char **sw, char **nw, char **ne ) {
  int away_type = b->uc;
  char *emblems[4] = { NULL, NULL, NULL, NULL };
  int i = 0;

  /* fixme: align this more closely with the microsoft icons */
  if ( b->present == GAIM_BUDDY_OFFLINE ) {
    emblems[i++] = "offline";
  } else if ( away_type == RVP_BUSY || away_type == RVP_PHONE ) {
    emblems[i++] = "occupied";
  } else if ( away_type != RVP_ONLINE ) {
    emblems[i++] = "away";
  }

  *se = emblems[0];
  *sw = emblems[1];
  *nw = emblems[2];
  *ne = emblems[3];
}

static char *rvp_status_text( GaimBuddy *b ) {
  if ( b->uc & UC_UNAVAILABLE ) {
    return g_strdup( "Unavailable" );
  }

  return NULL;
}

static void rvp_close( GaimConnection *gc ) {
  RVPData *rd = gc->proto_data;

  gaim_debug_misc( __FUNCTION__, "enter\n" );

#if 0
  /* currently unimplemented due to it dumping core at every possible
     opportunity. */
  if ( GAIM_CONNECTION_IS_CONNECTED( gc )) {
    /* oog. we need to block on this, really. */
    rd->me.unsubbed = FALSE;
    rvp_send_request( gc, "UNSUBSCRIBE", NULL );
    while( gtk_main_iteration_do( FALSE )) {
      if ( rd->me.unsubbed ) {
        break;
      }
    }
  }
#endif

  /* shut down the listener */
  if ( rd != NULL ) {
    if ( rd->linpa ) {
      gaim_input_remove( rd->linpa );
    }
    rd->linpa = 0;

    if ( rd->ld > 0 ) {
      close( rd->ld );
    }
    rd->ld = 0;
  }

  if ( rd->me.principal != NULL ) {
    g_free( rd->me.principal );
  }

  gc->proto_data = NULL;

  gaim_debug_misc( __FUNCTION__, "exit\n" );
}

/*
 * keepalive to monitor state of various subscriptions
 * currently does nothing.
 */
static void rvp_keepalive( GaimConnection *gc ) {
  struct rvp_data *rd = gc->proto_data;

  if ( rd->view_expiry - time( NULL ) < 120 ) {
    gaim_debug_misc( __FUNCTION__, "view expires in %d seconds, renewing\n",
                     rd->view_expiry - time( NULL ));
    rvp_send_request( gc, "PROPPATCH", NULL );
  }

  /* two minutes. windows client is at least this antsy */
  if ( rd->me.expires - time( NULL ) < 120 ) {
    GaimBuddyList *blist;
    GaimBlistNode *group, *cnode, *bnode;
    GaimBuddy *buddy;

    gaim_debug_misc( __FUNCTION__,
                     "main sub expires in %d seconds, renewing\n",
                     rd->me.expires - time( NULL ));

    /* resubscribe:
       send a SUBSCRIBE for me, and one for everything I'm subscribed
       to, and instead of a Notification-Type header use the existing
       subscription ID. */
    blist = gaim_get_blist();
    for ( group = blist->root; group; group = group->next ) {
      for ( cnode = group->child; cnode; cnode = cnode->next ) {
        if ( !GAIM_BLIST_NODE_IS_CONTACT(cnode)) {
          continue;
        }
        for ( bnode = cnode->child; bnode; bnode = bnode->next ) {
          if (!GAIM_BLIST_NODE_IS_BUDDY( bnode )) {
            continue;
          }
          buddy = (GaimBuddy *)bnode;
          serv_add_buddy( gc, buddy );
        }
      }
    }
    rvp_send_request( gc, "SUBSCRIBE", NULL );
  }
}

/*
 * ACL routines - currently non-functional xxx
 */
static void rvp_set_permit_deny(GaimConnection *gc) {
  gaim_debug_misc( __FUNCTION__, "RVP: Set permit deny\n");
}


static void rvp_add_permit(GaimConnection *gc, const char *who) {
  gaim_debug_misc( __FUNCTION__, "RVP: Add permit %s\n", who);
}


static void rvp_rem_permit(GaimConnection *gc, const char *who) {
  gaim_debug_misc( __FUNCTION__, "RVP: Rem permit %s\n", who);
}


static void rvp_add_deny(GaimConnection *gc, const char *who) {
  gaim_debug_misc( __FUNCTION__, "not implemented\n" );
}


static void rvp_rem_deny(GaimConnection *gc, const char *who) {
  gaim_debug_misc( __FUNCTION__, "not implemented\n" );
}

/*
 * list of supported away states
 */
static GList *rvp_away_states(GaimConnection *gc) {
  GList *m = NULL;

  m = g_list_append( m, _("Online"));
  m = g_list_append( m, _("Busy"));
  m = g_list_append( m, _("Idle"));
  m = g_list_append( m, _("Be Right Back"));
  m = g_list_append( m, _("Away From Computer"));
  m = g_list_append( m, _("On The Phone"));
  m = g_list_append( m, _("Out To Lunch"));
  m = g_list_append( m, _("Offline"));
  m = g_list_append( m, _("Appear Offline"));

  return m;
}

/*
 * mark me as being away
 */
static void rvp_set_away( GaimConnection *gc, const char *state, const char *msg ) {
  struct rvp_data *rd = gc->proto_data;
  char *away;
  gchar *args[1];

  gc->away = NULL;

  if (msg) {
    gc->away = "";
    away = "away";
    rd->self_away = TRUE;
    if ( rd->away_msg && ( strlen( rd->away_msg ) > 0))
      g_free( rd->away_msg );
    rd->away_msg = g_strdup( msg );
  } else if ( state ) {
    gc->away = "";

    if (!strcmp(state, "Away From Computer"))
      away = "away";
    else if (!strcmp(state, "Be Right Back"))
      away = "back-soon";
    else if (!strcmp(state, "Busy"))
      away = "busy";
    else if (!strcmp(state, "On The Phone"))
      away = "on-phone";
    else if (!strcmp(state, "Out To Lunch"))
      away = "at-lunch";
    else if (!strcmp(state, "Appear Offline"))
      away = "offline";
    else {
      gc->away = NULL;
      away = "online";
      rd->self_away = FALSE;
    }
  } else if ( gc->is_idle )
    away = "idle";
  else
    away = "online";

  if ( rd->away_stat )
    g_free( rd->away_stat );

  args[ 0 ] = away;
  rvp_send_request( gc, "PROPPATCH", args );

  rd->is_away_ok = FALSE;

  return;
}

/*
 * automatically set away when idle
 */
static void rvp_set_idle( GaimConnection *gc, int idle ) {
  struct rvp_data *rd = gc->proto_data;
  char *away;

  if ( rd->is_away_ok == FALSE )
    return;

  if ( gc->away )
    return;
  if ( idle )
    away = "idle";
  else
    away = "online";

  gaim_debug_misc( __FUNCTION__, "do we need to do something here?" );
}

/*
 * This is invoked when you context-click on a buddy. Its job is to
 * return menu items for the buddy. At present it doesn't appear to work
 * and we don't do anything useful with it anyway.
 */
static GList *rvp_buddy_menu( GaimBlistNode *node ) {
  GList *m = NULL;

  if ( GAIM_BLIST_NODE_IS_BUDDY(node)) {
    gaim_debug_misc( __FUNCTION__, "blart\n" );
  } else {
    gaim_debug_misc( __FUNCTION__, "not a buddy?\n" );
  }

  return m;
}

/*
 * set the state of a buddy
 */
static void rvp_set_buddy_status( GaimConnection *gc, gchar *nick, char *status ) {
  GaimAccount *ac = gaim_connection_get_account( gc );
  int online, stat;
  GaimBuddy *buddy;

  if ( nick == NULL ) { /* wtf? */
    gaim_debug_misc( __FUNCTION__, "ignoring null nick\n" );
    return;
  }

  buddy = gaim_find_buddy( ac, nick );

  if ( buddy == NULL ) {
    /* ain't no buddy of mine! */
    gaim_debug_misc( __FUNCTION__, "non-buddy %s is %s\n", nick, status );
    return;
  }

  gaim_debug_misc( __FUNCTION__, "%s is %s\n", buddy->name, status );

  online = 1;

  /* set buddy online status */
  if (!strcmp(status, "online")) {
    stat = RVP_ONLINE;
  } else if (!strcmp(status, "offline") || !strcmp(status, "invisible")) {
    online = 0;
    stat = RVP_OFFLINE;
  } else if (!strcmp(status, "idle")) {
    /*  stat |= UC_UNAVAILABLE | (RVP_IDLE << 1); */
    stat = RVP_IDLE;
  } else if (!strcmp(status, "away")) {
    stat = RVP_AWAY;
  } else if (!strcmp(status, "busy")) {
    stat = RVP_BUSY;
  } else if (!strcmp(status, "back-soon")) {
    stat = RVP_BRB;
  } else if (!strcmp(status, "on-phone")) {
    stat = RVP_PHONE;
  } else if (!strcmp(status, "at-lunch")) {
    stat = RVP_LUNCH;
  } else {
    /* catch new status and pop up a message window */
    gaim_notify_error(gc, _("Unknown status"), status, NULL );
    stat = RVP_UNKNOWN;
    return;
  }

  serv_got_update(gc, buddy->name, online, 0, 0, 0, stat );
}

/*
 * add a buddy
 */
static void rvp_add_buddy( GaimConnection *gc, GaimBuddy *buddy,
                           GaimGroup *group ) {
  RVPData *rd = gc->proto_data;
  GaimBuddy *dup;
  RVPBuddy *rb;
  gchar *args[4];

  rb = buddy->proto_data;

  /* new async code. yay! */
  if ( rb == NULL ) {
    rb = g_new0( RVPBuddy, 1 );
    rb->buddy = buddy;
    buddy->proto_data = rb;
  }

  if ( rb->principal == NULL ) {
    /* a bit overkill */
    GaimBuddy *temp = rvp_buddy_from_address( gc, buddy->name );

    if ( temp != NULL ) {
      rb->principal = g_strdup( ((RVPBuddy *)temp->proto_data)->principal );
      buddy->name = g_strdup( temp->name );

      rvp_free_buddy( temp );
    }
  }

  dup = rvp_get_temp_buddy( gc, buddy );
  if ( dup ) {
    RVPBuddy *drb = dup->proto_data;
    /* copy what we've got into the 'real' buddy structure, and delete
       the temp buddy */
    g_hash_table_remove( rd->nonbuddy, rb->principal );

    /* move the RVP data into place */
    g_free( rb );
    buddy->proto_data = drb;
    drb->buddy = buddy;
    rb = drb;

    rvp_dump_buddy( buddy );
  }

  /* don't subscribe if we're not connected (duh) */
  if ( GAIM_CONNECTION_IS_CONNECTED( gc )) {
    xmlDocPtr doc;
    xmlNodePtr root, acl, ace, node1, node2, node3;
    xmlChar *content;
    gint content_length;

    args[0] = (gchar *)buddy;
    args[1] = NULL;
    rvp_send_request( gc, "SUBSCRIBE", args );

    /* We only need to send an ACL if we don't already have an
       appropriate one set. send-to and presence are sufficient for a
       buddy, plus at least one of assertion, digest or ntlm */
    if (( rb->acl.credentials.assertion ||
          rb->acl.credentials.digest ||
          rb->acl.credentials.ntlm ) &&
        rb->acl.send_to == TRUE && rb->acl.presence == TRUE ) {
      return;
    }

    doc = xmlNewDoc( BAD_CAST "1.0" );
    root = xmlNewNode( NULL, BAD_CAST "r:rvpacl" );
    xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
              BAD_CAST "r" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
              BAD_CAST "a" );
    xmlDocSetRootElement( doc, root );

    acl = xmlNewNode( NULL, BAD_CAST "a:acl" );
    node1 = xmlNewNode( NULL, BAD_CAST "a:inheritance" );
    node2 = xmlNewText( BAD_CAST "none" );
    xmlAddChild( node1, node2 );
    xmlAddChild( acl, node1 );

    ace = xmlNewNode( NULL, BAD_CAST "a:ace" );

    node1 = xmlNewNode( NULL, BAD_CAST "a:principal" );
    node2 = xmlNewNode( NULL, BAD_CAST "a:rvp-principal" );
    node3 = xmlNewText( BAD_CAST rb->principal );
    xmlAddChild( node2, node3 );
    xmlAddChild( node1, node2 );
    xmlAddChild( ace, node1 );

    node2 = xmlNewNode( NULL, BAD_CAST "a:credentials" );
    node3 = xmlNewNode( NULL, BAD_CAST "a:assertion" );
    xmlAddChild( node2, node3 );
    node3 = xmlNewNode( NULL, BAD_CAST "a:digest" );
    xmlAddChild( node2, node3 );
    node3 = xmlNewNode( NULL, BAD_CAST "a:ntlm" );
    xmlAddChild( node2, node3 );
    xmlAddChild( node1, node2 );
    xmlAddChild( ace, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "a:grant" );
    node2 = xmlNewNode( NULL, BAD_CAST "a:send-to" );
    xmlAddChild( node1, node2 );
    node2 = xmlNewNode( NULL, BAD_CAST "a:presence" );
    xmlAddChild( node1, node2 );
    xmlAddChild( ace, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "a:deny" );
    xmlAddChild( ace, node1 );
    xmlAddChild( acl, ace );
    xmlAddChild( root, acl );

    xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
    xmlFreeDoc( doc );

    args[0] = g_strdup_printf( rb->principal );
    args[1] = (gchar *)content;
    args[2] = (gchar *)content_length;
    args[3] = NULL;

    rvp_send_request( gc, "ACL", args );

    g_free( args[0] );
    /* g_free( content ); DON'T DO THIS! It's freed later. */
  } else {
    gaim_debug_warning( __FUNCTION__, "not connected\n" );
  }
}

/*
 * remove a buddy from the buddy list
 */
static void rvp_rem_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group) {
  gchar *args[1];

  gaim_debug_misc( __FUNCTION__, "rem_buddy name: %s\n", buddy->name );

  args[0] = (gchar *)buddy;

  rvp_send_request( gc, "UNSUBSCRIBE", args );
}

/*
 * generic notify function
 */
static int rvp_send_notify( GaimConnection *gc, const char *who, gint msgtype,
                            gchar *message, void *data ) {
  int typing = FALSE;
  GaimConversation *conv = NULL;
  GaimConvImFlags flags;
  GaimAccount *ac = gaim_connection_get_account( gc );
  GList *recipients = NULL, *r;
  GaimBuddy *buddy;
  RVPBuddy *rb = NULL;
  RVPData *rd = gc->proto_data;
  const gchar *fullname;
  gchar *msg = NULL, *sessid = NULL;
  gchar *args[3];
  xmlDocPtr doc;
  xmlNodePtr root, xmlmsg, node, xmlcontact, body, mime;
  xmlChar *content = NULL;
  gint content_length;
  GList *buddies;

  switch( msgtype ) {
  case RVP_MSG_TYPING:
    typing = *(int*)data;
    break;

  case RVP_MSG_IM:
    flags = *(GaimConvImFlags*)data;
    break;

  case RVP_MSG_INVITE:
  case RVP_MSG_CHAT:
  case RVP_MSG_LEAVE:
    conv = (GaimConversation *)data;
    sessid = gaim_conversation_get_data( conv, "sessid" );
    break;

  default:
    break;
  }

  if ( msgtype == RVP_MSG_TYPING && typing != TRUE ) {
    return EXCHANGE_TYPING_TIMEOUT;
  }

  if ( sessid == NULL || msgtype == RVP_MSG_INVITE ) {
    /* pick up a hostname if we don't have one */
    fullname = rvp_normalize( gaim_connection_get_account( gc ), who );

    /*
     * Is it someone we know about?
     */
    buddy = gaim_find_buddy( ac, who );
    if ( buddy == NULL ) {
      buddy = gaim_find_buddy( ac, fullname );
    }

    if ( buddy == NULL ) {
      GaimBuddy *gb = rvp_buddy_from_address( gc, (gchar *)fullname );
      /* step 2, is it someone we're already conversing with? */
      /* gaim_get_conversations returns a glist of conversations from
         which I should be able to go digging for people */

      if ( gb == NULL ) {
        gaim_debug_misc( __FUNCTION__, "rvp_buddy_from_address returned null for %s\n", fullname );
        return ( msgtype == RVP_MSG_TYPING ? EXCHANGE_TYPING_TIMEOUT : 0 );
      }

      buddy = rvp_get_temp_buddy_create( gc, gb );
    }

    rb = buddy->proto_data;
    if ( rb == NULL ) {
      gaim_debug_misc( __FUNCTION__, "you've not done the proto_data hookup\n" );
      return 0;
    }

    if ( rb->sessionid == NULL ) {
      rb->sessionid = rvp_get_sessid();
    }

    if ( sessid == NULL ) {
      sessid = rb->sessionid;
    }
  }

  switch( msgtype ) {
  case RVP_MSG_TYPING:
    msg = g_strdup_printf( "MIME-Version: 1.0\r\n"
                           "Content-Type: text/x-msmsgscontrol\r\n"
                           "TypingUser: %s\r\n"
                           "Session-Id: {%s}\r\n\r\n",
                           rd->me.buddy->name,
                           sessid );
    break;

  case RVP_MSG_INVITE:
    msg = g_strdup_printf( "MIME-Version: 1.0\r\n"
                           "Content-Type: text/x-msmsgscontrol\r\n"
                           "Session-Id: {%s}\r\n\r\n\r\n",
                           sessid );
    break;

  case RVP_MSG_IM:
  case RVP_MSG_CHAT:
    msg = g_strdup_printf( "MIME-Version: 1.0\r\n"
                           "Content-Type: text/plain; charset=UTF-8\r\n"
                           "X-MMS-IM-Format: FN=MS%%20Shell%%20Dlg; EF=; CO=0; CS=0; PF=0\r\n"
                           "Session-Id: {%s}\r\n\r\n%s",
                           sessid,
                           message );
    break;

  case RVP_MSG_LEAVE:
    msg = g_strdup_printf( "MIME-Version: 1.0\r\n"
                           "Content-Type: text/x-imleave\r\n"
                           "Session-Id: {%s}\r\n\r\n\r\n",
                           sessid );
    break;

  default:
    break;
  }

  /* xml content */
  doc = xmlNewDoc( BAD_CAST "1.0" );
  root = xmlNewNode( NULL, BAD_CAST "r:notification" );
  xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
  xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
            BAD_CAST "r" );
  xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
            BAD_CAST "a" );
  xmlDocSetRootElement( doc, root );

  xmlmsg = xmlNewNode( NULL, BAD_CAST "r:message" );

  /* sender */
  node = xmlNewNode( NULL, BAD_CAST "r:notification-from" );
  xmlcontact = build_contact( NULL, NULL, "", rd->me.principal );
  xmlAddChild( node, xmlcontact );
  xmlAddChild( xmlmsg, node );

  /* recipient */
  node = xmlNewNode( NULL, BAD_CAST "r:notification-to" );
  switch( msgtype ) {
  case RVP_MSG_TYPING:
  case RVP_MSG_IM:
    xmlcontact = build_contact( NULL, NULL, "", rb->principal );
    xmlAddChild( node, xmlcontact );
    gaim_debug_misc( __FUNCTION__, "recip: %s\n", rb->principal );
    recipients = g_list_append( recipients, g_strdup( rb->principal ));
    break;

  case RVP_MSG_INVITE:
  case RVP_MSG_CHAT:
  case RVP_MSG_LEAVE:
    if ( msgtype == RVP_MSG_INVITE ) {
      recipients = g_list_append( recipients, g_strdup( rb->principal ));
      xmlcontact = build_contact( NULL, NULL, "", rb->principal );
      xmlAddChild( node, xmlcontact );
    }
    /* who's all up in there? */
    buddies = gaim_conv_chat_get_users( GAIM_CONV_CHAT( conv ));
    while( buddies ) {
      GaimConvChatBuddy *cb = (GaimConvChatBuddy *)buddies->data;
      GaimBuddy *b;

      /* gah. chat only stores the username */
      if (( b = rvp_find_buddy( gc, cb->name, NULL, NULL )) == NULL ) {
        b = rvp_buddy_from_address( gc, cb->name ); /* leak */
      }

      rb = b->proto_data;
      xmlcontact = build_contact( NULL, NULL, "", rb->principal );
      xmlAddChild( node, xmlcontact );
      if ( msgtype != RVP_MSG_INVITE ) {
        recipients = g_list_append( recipients, g_strdup( rb->principal ));
      }
      buddies = buddies->next;
    }
    break;

  default:
    break;
  }

  xmlAddChild( xmlmsg, node );

  /* message */
  body = xmlNewNode( NULL, BAD_CAST "r:msgbody" );
  mime = xmlNewNode( NULL, BAD_CAST "r:mime-data" );

  node = xmlNewCDataBlock( doc, BAD_CAST msg, strlen( msg ));
  xmlAddChild( mime, node );
  xmlAddChild( body, mime );
  xmlAddChild( xmlmsg, body );
  xmlAddChild( root, xmlmsg );

  /* render it */
  xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
  xmlFreeDoc( doc );

  /* everyone needs a 303^WNOTIFY */
  for ( r = recipients; r != NULL; r = r->next ) {
    gaim_debug_misc( __FUNCTION__, "notify to %s\n", (gchar *)r->data );

    /* shouldn't happen! */
    if ( r->data == NULL ) {
      continue;
    }

    args[0] = (gchar *)r->data;
    /* this gets freed elsewhere, so we have to send a duplicate */
    args[1] = g_malloc0( content_length );
    memcpy( args[1], content, content_length );
    args[2] = (gchar *)content_length;

    rvp_send_request( gc, "NOTIFY", args );
  }
  g_free( content );
  g_list_free( recipients );

  if ( msgtype == RVP_MSG_TYPING ) {
    return EXCHANGE_TYPING_TIMEOUT;
  } else {
    return 1;
  }
}

/*
 * send a typing notification
 */
static int rvp_send_typing( GaimConnection *gc, const char *who, int typing ) {
  return rvp_send_notify( gc, who, RVP_MSG_TYPING, NULL, &typing );
}

/*
 * send an instant message
 */
static int rvp_send_im( GaimConnection *gc, const char *who,
                        const char *message, GaimConvImFlags flags ) {
  return rvp_send_notify( gc, who, RVP_MSG_IM, (gchar *)message, &flags );
}


/*
 * invite someone to a chat
 */
void rvp_chat_invite( GaimConnection *gc, int id, const char *msg, const char *who ) {
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return;
  }

  gaim_debug_misc( __FUNCTION__, "inviting %s to chat %d\n", who, id );

  rvp_send_notify( gc, who, RVP_MSG_INVITE, NULL, b );
  gaim_conv_chat_add_user( GAIM_CONV_CHAT( b ), who, NULL, GAIM_CBFLAGS_NONE,
                           TRUE );
}

/*
 * join a chat
 */
void rvp_chat_join( GaimConnection *gc, GHashTable *components ) {
  GaimConversation *conv = NULL;
  GaimBuddy *buddy = NULL;
  gchar *sessid = NULL;
  gint id = 0;
  RVPData *rd = gc->proto_data;

  if ( components != NULL ) {
    sessid = g_hash_table_lookup( components, "sessid" );
    buddy = g_hash_table_lookup( components, "from" );
  }

  /* self-created chat */
  if ( sessid == NULL ) {
    sessid = rvp_get_sessid();
  }
  if ( buddy == NULL ) {
    buddy = rd->me.buddy;
  }

  if ( rd->chats == NULL ) {
    rd->chats = g_hash_table_new( g_str_hash, g_str_equal );
    rd->chatid = 1;
  } else {
    conv = g_hash_table_lookup( rd->chats, sessid );
    if ( conv == NULL ) {
      id = rd->chatid;
    } else {
      id = gaim_conv_chat_get_id( GAIM_CONV_CHAT( conv ));
    }
  }

  if ( conv == NULL ) {
    /* not sure about the third param */
    conv = serv_got_joined_chat( gc, id, buddy->name );
  }
  g_hash_table_replace( rd->chats, g_strdup( sessid ), conv );
  gaim_debug_misc( __FUNCTION__, "created new chat %p\n", conv );
  gaim_conversation_set_data( conv, "sessid", sessid );
  gaim_conv_chat_add_user( GAIM_CONV_CHAT( conv ), buddy->name, NULL,
                           GAIM_CBFLAGS_NONE, TRUE );

#if 0
  /* add me also, unless I'm already there
     not so sure about this
  */
  buddies = gaim_conv_chat_get_users( GAIM_CONV_CHAT( conv ));
  if ( buddies ) {
    while( buddies ) {
      gaim_conv_chat_add_user( GAIM_CONV_CHAT( conv ), rd->me.buddy->name,
                               NULL, GAIM_CBFLAGS_NONE, TRUE );
    }
  }
#endif
}

void rvp_chat_leave( GaimConnection *gc, int id ) {
  RVPData *rd = gc->proto_data;
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return;
  }

  gaim_debug_misc( __FUNCTION__, "leaving conv %p\n", b );

  rvp_send_notify( gc, rd->me.buddy->name, RVP_MSG_LEAVE, NULL, b );
}

void rvp_chat_user_left( GaimConnection *gc, int id, GaimBuddy *buddy ) {
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return;
  }

  gaim_conv_chat_remove_user( GAIM_CONV_CHAT(b), buddy->name, NULL );
}

int rvp_chat_send( GaimConnection *gc, int id, const char *message ) {
  RVPData *rd = gc->proto_data;
  GSList *bcs = gc->buddy_chats;
  GaimConversation *b = NULL;

  while( bcs ) {
    b = (GaimConversation *)bcs->data;
    if ( id == gaim_conv_chat_get_id( GAIM_CONV_CHAT( b ))) {
      break;
    }
    b = NULL;
    bcs = bcs->next;
  }

  if ( !b ) {
    return 0;
  }

  gaim_debug_misc( __FUNCTION__, "sending message to conv %p\n", b );

  return rvp_send_notify( gc, rd->me.buddy->name, RVP_MSG_CHAT,
                          (gchar *)message, b );
}

/*
 * set up the listener for async notification
 */
void rvp_setup_listener( GaimConnection *gc ) {
  struct rvp_data *rd = gc->proto_data;
  struct sockaddr_in name;
  struct sockaddr_in *ptr;
  struct sockaddr addr;
  int ld;
  socklen_t addrlen;
  int rc;

  /* verify that this error works... it's being called from inside a
     callback, so it may not. */
  if (( ld = socket( AF_INET, SOCK_STREAM, 0)) < 0 ) {
    gaim_connection_error(gc, _("Unable to create listener"));
    gc->wants_to_die = TRUE;
    return;
  }

  fcntl(ld, F_SETFL, O_NONBLOCK);

  name.sin_family = AF_INET;

  /* Initialize port value to 0 (let the system pick a port).
   * If the user has specified a fixed_port value, and it
   * looks like a valid number, we use that instead.
   */
  name.sin_port = 0;
  if ( rd->listenport > 1023 ) {
    name.sin_port = htons( rd->listenport );
  } else {
    if ( rd->listenport != 0 ) {
      gaim_debug_warning( __FUNCTION__, "ignoring privileged fixed port %d\n",
                          rd->listenport);
    }
  }

  name.sin_addr.s_addr = INADDR_ANY;

  if ( bind( ld, (struct sockaddr *)&name, sizeof( name ) ) < 0 ) {
    close( ld );

    /* Verify, as above */
    gaim_connection_error(gc, _("Unable to bind listener"));
    gc->wants_to_die = TRUE;
    return;
  }

  /* retrieve the port we've set up on */
  addrlen = sizeof( addr );
  if (( rc = getsockname( ld, &addr, &addrlen )) < 0) {
    gaim_connection_error(gc, _("Unable to get socket name"));
    gc->wants_to_die = TRUE;
    return;
  }

  ptr = (struct sockaddr_in *)&addr;
  rd->lport = ntohs( ptr->sin_port );
  gaim_debug_misc( __FUNCTION__, "listener port: %d\n", rd->lport, ptr->sin_port );

  if ( listen( ld, 5 ) < 0 ) {
    close( ld );
    gaim_connection_error(gc, _("Unable to set listener"));
    gc->wants_to_die = TRUE;
    return;
  }

  rd->ld = ld;
}

/*
 * this is largely cribbed from Gaim's util.c
 */
static void
destroy_fetch_url_data(GaimFetchUrlData *gfud) {
  if (gfud->request.webdata         != NULL) g_free(gfud->request.webdata);
  if (gfud->request.header          != NULL) g_free(gfud->request.header);
  if (gfud->response.webdata         != NULL) g_free(gfud->response.webdata);
  if (gfud->response.header          != NULL) g_free(gfud->response.header);
  if (gfud->url             != NULL) g_free(gfud->url);
  if (gfud->method          != NULL) g_free(gfud->method);
  if (gfud->website.address != NULL) g_free(gfud->website.address);
  if (gfud->website.page    != NULL) g_free(gfud->website.page);
  if (gfud->website.user    != NULL) g_free(gfud->website.user);
  if (gfud->website.passwd  != NULL) g_free(gfud->website.passwd);
  g_free(gfud);
}

static size_t parse_content_len( const char *data, size_t data_len ) {
  size_t content_len = 0;
  const char *p = NULL;

  /* This is still technically wrong, since headers are
   * case-insensitive [RFC 2616, section 4.2], though this ought to
   * catch the normal case.  Note: data is _not_ nul-terminated.
   */
  if ( data_len > 16 ) {
    p = strncasecmp( data, "Content-Length: ", 16 ) == 0 ? data : NULL;
    if ( !p ) {
      p = g_strstr_len( data, data_len, "\nContent-Length: " );
      if ( p )
        p += 1;
    }
  }

  if ( p != NULL && g_strstr_len( p, data_len - ( p - data ), "\n")) {
    /* this appears to make valgrind unhappy */
    content_len = atoi( &p[strlen( "Content-Length: " )]);
  }

  return content_len;
}

static void url_fetched_cb( gpointer url_data, gint sock,
                            GaimInputCondition cond ) {
  GaimFetchUrlData *gfud = url_data;
  char data;
  gboolean got_eof = FALSE;

  gfud->sock = sock;
  if (sock == -1) {
    gfud->callback(gfud->user_data, gfud, 0);
    if ( !gfud->preserve ) {
      destroy_fetch_url_data(gfud);
    }
    return;
  }

 resend:
  if (!gfud->sentreq) {
    char *buf;
    gsize content_length = 0;

    gfud->resend = FALSE;

    /* we need to encode the outgoing buffer in the Microsofty way */
    if ( gfud->request.webdata && !gfud->encoded ) {
      GError *error = NULL;
      gsize read, i;
      gchar *bigbuf = NULL;
      gchar *converted = NULL;

      if ( gfud->request.length ) {
        read = gfud->request.length;
      } else {
        gaim_debug_misc( __FUNCTION__, "no length set, you fule\n" );
        read = strlen( gfud->request.webdata );
        gfud->request.length = read;
      }

      bigbuf = g_malloc0( read * 2 );

      for ( i = 0; i < read; i++ ) {
        bigbuf[ i * 2 ] = gfud->request.webdata[ i ];
      }

      converted = g_convert( bigbuf, gfud->request.length * 2,
                             "UTF-8", "UCS-2LE", &read,
                             &content_length, &error );
      if ( !converted ) {
        /* xxx */
        gaim_debug_misc( __FUNCTION__, "conversion error\n" );
      } else {
        g_free( gfud->request.webdata );
        gfud->request.webdata = g_realloc( converted, content_length + 1 );
        gfud->request.webdata[ content_length ] = '\0';
      }
    } else {
      content_length = gfud->request.length;
    }

    gfud->encoded = TRUE;

    buf = g_strdup_printf( "%s %s%s HTTP/1.1\r\n"
                           "Content-Length: %u\r\n"
                           "RVP-Notifications-Version: 0.2\r\n"
                           "Host: %s\r\n%s\r\n%s" ,
                           gfud->method,
                           (gfud->full ? "" : "/"),
                           (gfud->full ? gfud->url : gfud->website.page),
                           content_length,
                           gfud->website.address,
                           gfud->request.header ? gfud->request.header : "",
                           gfud->request.webdata ?
                           gfud->request.webdata : "" );

    gaim_debug_misc( "gaim_url_fetch", "Requesting %p:\n%s\n", gfud, buf );

    write( sock, buf, strlen( buf ));
    fcntl( sock, F_SETFL, O_NONBLOCK );
    gfud->sentreq = TRUE;
    gfud->inpa = gaim_input_add(sock, GAIM_INPUT_READ,
                                url_fetched_cb, url_data);
    gfud->data_len = 4096;
    gfud->response.webdata = g_malloc(gfud->data_len);

    return;
  }

  if (read(sock, &data, 1) > 0 || errno == EWOULDBLOCK) {
    if (errno == EWOULDBLOCK) {
      errno = 0;
      return;
    }

    gfud->len++;

    if (gfud->len == gfud->data_len + 1) {
      gfud->data_len += (gfud->data_len) / 2;

      gfud->response.webdata =
        g_realloc(gfud->response.webdata, gfud->data_len);
    }

    gfud->response.webdata[gfud->len - 1] = data;

    if (!gfud->startsaving) {
      if (data == '\r')
        return;

      if (data == '\n') {
        if (gfud->newline) {
          size_t content_len;
          gfud->startsaving = TRUE;

#ifdef LOUD
          gaim_debug_misc( __FUNCTION__, "Response headers:\n%*.*s\n",
                           gfud->len, gfud->len, gfud->response.webdata );
#else
          gaim_debug_misc( __FUNCTION__, "got headers\n" );
#endif

          /* See if we can find a content length. */
          content_len = parse_content_len(gfud->response.webdata, gfud->len);

          if (content_len == 0) {
            /* We'll stick with an initial 8192 */
            content_len = 8192;
          } else {
            gfud->has_explicit_data_len = TRUE;
          }

          /* Out with the old... */
          gfud->response.header = gfud->response.webdata;
          gfud->response.header =
            g_realloc( gfud->response.header, gfud->len + 1 );
          gfud->response.header[gfud->len] = 0;
          gfud->len = 0;
          gfud->response.webdata = NULL;

          /* In with the new. */
          gfud->data_len = content_len;
          gfud->response.webdata = g_try_malloc(gfud->data_len);
          if (gfud->response.webdata == NULL) {
            gaim_debug_error( "gaim_url_fetch", "Failed to allocate %u bytes: %s\n", gfud->data_len, strerror( errno ));
            gaim_input_remove( gfud->inpa );
            gfud->callback(gfud->user_data, gfud, 0);

            if ( gfud->resend ) {
              goto resend;
            }

            close(sock);
            if ( !gfud->preserve ) {
              destroy_fetch_url_data(gfud);
            }
          }
        } else {
          gfud->newline = TRUE;
        }

        return;
      }

      gfud->newline = FALSE;
    } else if (gfud->has_explicit_data_len && gfud->len == gfud->data_len) {
      got_eof = TRUE;
    }
  } else if (errno != ETIMEDOUT) {
    got_eof = TRUE;
  } else {
    gaim_input_remove(gfud->inpa);
    gfud->callback(gfud->user_data, gfud, 0);

    if ( gfud->resend ) {
      goto resend;
    }

    close(sock);

    if ( !gfud->preserve ) {
      destroy_fetch_url_data( gfud );
    }
  }

  if (got_eof) {
    gfud->response.webdata = g_realloc(gfud->response.webdata, gfud->len + 1);
    gfud->response.webdata[gfud->len] = 0;
    gaim_input_remove(gfud->inpa);
    gfud->callback(gfud->user_data, gfud, gfud->len);
    if ( gfud->resend ) {
      goto resend;
    }
    close(sock);

    if ( !gfud->preserve ) {
      destroy_fetch_url_data( gfud );
    }
  }
}

/* cribbed from LWP::Authen::Digest */
/* CAUTION currently does not handle qop */
static gchar *get_auth_digest( GaimFetchUrlData *gfud, gchar *header,
                               gchar *user, gchar *pass ) {
  gchar **auth_param = NULL;
  gint i = 0;
  gchar *nonce = NULL, *realm = NULL;
  gchar *retval = NULL, hexresp[33];
  md5_state_t md5;
  md5_byte_t seekrit1[16], seekrit2[16], response[16];
  gchar *uri;

  /* slightly naive */
  gaim_debug_misc( __FUNCTION__, "parsing header %s\n", header );
  auth_param = g_strsplit( &header[strlen( "Digest ")], ", ", 0 );

  /* find the nonce */
  for ( i = 0; auth_param[i]; i++ ) {
    gaim_debug_misc( __FUNCTION__, "%d: %s\n", i, auth_param[i] );
    if ( g_str_has_prefix( auth_param[i], "nonce=\"" )) {
      nonce = g_strdup( &auth_param[i][7] );
      nonce[strlen( nonce ) - 1] = '\0';
    } else if ( g_str_has_prefix( auth_param[i], "realm=\"" )) {
      realm = g_strdup( &auth_param[i][7] );
      realm[strlen( realm ) - 1] = '\0';
    }
  }

  if ( nonce == NULL || realm == NULL ) {
    gaim_debug_error( __FUNCTION__,
                      "Can't find one of nonce or realm in header\n" );
    return NULL;
  }

  if ( gfud->website.page[0] != '/' ) {
    uri = g_strconcat( "/", gfud->website.page, NULL );
  } else {
    uri = g_strdup( gfud->website.page );
  }

  md5_init( &md5 );
  md5_append( &md5, (md5_byte_t *)user, strlen( user ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)realm, strlen( realm ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)pass, strlen( pass ));
  md5_finish( &md5, seekrit1 );

  /* digest = ( seekrit1 ) */
  /* digest = ( seekrit1, nonce ) */

  md5_init( &md5 );
  md5_append( &md5, (md5_byte_t *)gfud->method, strlen( gfud->method ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)uri, strlen( uri));
  md5_finish( &md5, seekrit2 );

  /* digest = ( seekrit1, nonce, seekrit2 ) */

  md5_init( &md5 );
  for ( i = 0; i < 16; i++ ) {
    gchar *hex = g_strdup_printf( "%02x", seekrit1[i] );
    md5_append( &md5, (md5_byte_t *)hex, 2 );
    g_free( hex );
  }
  md5_append( &md5, (md5_byte_t *)":", 1 );
  md5_append( &md5, (md5_byte_t *)nonce, strlen( nonce ));
  md5_append( &md5, (md5_byte_t *)":", 1 );
  for ( i = 0; i < 16; i++ ) {
    gchar *hex = g_strdup_printf( "%02x", seekrit2[i] );
    md5_append( &md5, (md5_byte_t *)hex, 2 );
    g_free( hex );
  }
  md5_finish( &md5, response );

  for ( i = 0; i < 16; i++ ) {
    sprintf( &hexresp[i * 2], "%02x", response[i] );
  }
  hexresp[32] = '\0';

  retval = g_strdup_printf( "Digest "
                            "username=\"%s\", "
                            "realm=\"%s\", "
                            "algorithm=\"%s\", "
                            "uri=\"%s\", "
                            "nonce=\"%s\", "
                            "response=\"%s\"",
                            user,
                            realm,
                            "MD5",
                            uri,
                            nonce,
                            hexresp );

  if ( auth_param != NULL ) {
    g_strfreev( auth_param );
  }

  return retval;
}

/*
 * handle data from async notifications
 */
static void rvp_async_data( void *udata, GaimFetchUrlData *gfd, size_t size ) {
  GaimConnection *gc = udata;
  GaimAccount *ac = gaim_connection_get_account( gc );
  RVPData *rd = gc->proto_data;
  size_t i;
  const char *p = NULL;
  gboolean is_xml = FALSE;
  gint code = 0;

  /* Oh KWATZ. */
  gaim_debug_misc( __FUNCTION__, "callback on gfud %p\n", gfd );

  /* to be sure this gets done right... */
  gfd->preserve = FALSE;

  /* don't bother continuing if the connection is shutting down. */
  if ( gc->wants_to_die ) {
    return;
  }

  /* Rare, but can happen */
  if ( rd == NULL ) {
    return;
  }

  if ( gfd->response.header == NULL ) {
    gchar *msg = g_strdup_printf( _("Server returned empty response to %s request"),
                                  gfd->method ? gfd->method : _("unknown"));
    gaim_connection_error( gc, msg );
    g_free( msg );

    if ( gfd->method != NULL ) {
      if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
        gaim_debug_warning( __FUNCTION__, "dying...\n" );
        gc->wants_to_die = TRUE;
      }
    } else {
      /* connection still dies when this happens, and I'm not clear
         why */
      gaim_debug_error( __FUNCTION__, "Someone's probably hacking you...\n" );
    }
    return;
  }

  if ( gfd->method != NULL ) {
    /* find out what the status code is */
    if (( !strncmp( gfd->response.header, "HTTP/1.1 ", 9 )) &&
        strlen( gfd->response.header ) >= 12 &&
        ( code = atoi( &gfd->response.header[9] )) > 99 && code < 1000 ) {
      gaim_debug_misc( __FUNCTION__, "got a %d response to %p %s %s\n", code,
                       gfd, gfd->method, gfd->website.page );
    } else {
      gaim_connection_error( gc, _("Unexpected response from server" ));
      return;
    }
  } else {
    gaim_debug_misc( __FUNCTION__, "async notification\n" );
  }

  if ( gfd->response.header == NULL ) {
      gaim_connection_error( gc, _("No headers in response from server" ));
      return;
  }

  /* redirect */
  if ( code == 302 ) {
    gchar *newurl = get_header_content( gfd->response.header, "Location" );
    if ( newurl != NULL ) {
      gaim_debug_warning( __FUNCTION__, "untested redirection code\n" );
      g_free( gfd->response.header );
      gfd->response.header = NULL;
      gfd->len = 0;
      gfd->sentreq = FALSE;
      gfd->startsaving = FALSE;
      gfd->newline = FALSE;
      gfd->resend = TRUE;

      /* for now, this only handle internal redirects because of the
         Host header xxx */

      /* In theory, we'd patch the buddy that caused the redirect so
         that future messages go to the correct URL. In practice, this
         appears to be a bad idea, so I've not done it. */

      if ( gfd->url             != NULL) g_free( gfd->url );
      if ( gfd->website.address != NULL) g_free( gfd->website.address );
      if ( gfd->website.page    != NULL) g_free( gfd->website.page );
      if ( gfd->website.user    != NULL) g_free( gfd->website.user );
      if ( gfd->website.passwd  != NULL) g_free( gfd->website.passwd );
      gfd->url = g_strdup( newurl );
      gfd->tried_auth = FALSE; /* hmm */
      gaim_url_parse( gfd->url, &gfd->website.address, &gfd->website.port,
                      &gfd->website.page, &gfd->website.user,
                      &gfd->website.passwd );
      return;
    }
    /* xxx what's to do here? */
    gaim_connection_error( gc, _( "Got a redirect with no location" ));
    return;
  }

  /* look for an auth failure */
  if ( code == 401 ) {
    gchar *newheaders = NULL;
    gchar *authtype = NULL;

    if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
      gaim_connection_update_progress( gc, _("Authenticating"), rd->login_step,
                                       MAX_LOGIN_STEPS );
    }

    /* headers[0] is the HTTP response line */
    authtype = get_header_content( gfd->response.header, "WWW-Authenticate" );
    if ( authtype == NULL ) {
      gaim_connection_error( gc, _("Auth error, but no auth type requested" ));
      return;
    }

    if ( !strcmp( authtype, "NTLM" )) {
      gchar *ntlm = get_ntlm_msg1( rd->domain, rd->client_host );
      rd->auth_type = RVP_AUTH_NTLM;
      newheaders = g_strdup_printf( "Authorization: %s\r\n", ntlm );
      /* gfd->tried_auth = FALSE; HMMM. Fixme */
    } else if ( g_str_has_prefix( authtype, "NTLM " )) {
      gchar *ntlm;

      if ( gfd->tried_auth ) {
        gaim_debug_warning( __FUNCTION__, "tried authing already!\n" );
        /* prompt for a password! */
        gaim_connection_error( gc, _("Auth failed"));
        ac->password = NULL;
        gc->wants_to_die = TRUE;
        return;
      }

      gfd->tried_auth = TRUE;

      ntlm = get_auth_ntlm( authtype, rd->user, ac->password,
                            rd->client_host, rd->domain,
                            rd->authid );

      newheaders = g_strdup_printf( "Authorization: %s\r\n", ntlm );
    } else if ( g_str_has_prefix( authtype, "Digest" )) {
      gchar *digest = NULL;

      if ( gfd->tried_auth ) {
        /* prompt for a password! */
        gaim_connection_error( gc, _("Auth failed"));
        ac->password = NULL;
        gc->wants_to_die = TRUE;
        return;
      }
      gfd->tried_auth = TRUE;

      digest = get_auth_digest( gfd, authtype, rd->user, ac->password );

      newheaders = g_strdup_printf( "Authorization: %s\r\n", digest );
    } else {
      gaim_connection_error( gc, _("unrecognised auth type"));
      gaim_debug_error( __FUNCTION__, "unrecognised auth type %s\n",
                        authtype );
      gc->wants_to_die = TRUE;
      return;
    }

    /* a bit silly, esp since newheaders MUST NOT be NULL */
    if ( newheaders != NULL ) {
      gchar **oh = g_strsplit( gfd->request.header, "\r\n", 0 );
      gchar **nh = g_strsplit( newheaders, "\r\n", 0 );
      gchar *newh = NULL, *build = NULL;
      int j;

      if ( nh != NULL && oh != NULL ) {
        for ( j = 0; oh[j] != NULL ; j++ ) {
          gchar **bits = g_strsplit( oh[j], ":", 2 ); /* overkill */
          gchar *hdr = oh[j];

          if ( bits[0] == NULL ) {
            continue;
          }

          for ( i = 0; nh[i]; i++ ) {
            gchar **b2 = g_strsplit( nh[i], ":", 2 );
            if ( b2[0] == NULL ) {
              continue;
            }
            if ( !strcmp( b2[0], bits[0] )) {
              hdr = nh[i];
            }
            g_strfreev( b2 );
          }

          if ( hdr && strlen( hdr )) {
            if ( newh == NULL ) {
              newh = g_strdup_printf( "%s\r\n", hdr );
            } else {
              build = g_strdup_printf( "%s%s\r\n", newh, hdr );
              g_free( newh );
              newh = build;
            }
          }

          g_strfreev( bits );
        }

        for ( i = 0; nh[i]; i++ ) {
          gchar **b2 = g_strsplit( nh[i], ":", 2 );
          gchar *hdr = nh[i];

          for ( j = 0; oh[j]; j++ ) {
            gchar **bits = g_strsplit( oh[j], ":", 2 ); /* overkill */

            if ( bits[0] == NULL ) {
              continue;
            }

            if ( b2[0] == NULL ) {
              continue;
            }
            if ( !strcmp( b2[0], bits[0] )) {
              hdr = NULL;
            }
            g_strfreev( bits );
          }

          if ( hdr && strlen( hdr )) {
            if ( newh == NULL ) {
              newh = g_strdup_printf( "%s\r\n", hdr );
            } else {
              build = g_strdup_printf( "%s%s\r\n", newh, hdr );
              g_free( newh );
              newh = build;
            }
          }

          g_strfreev( b2 );
        }

        g_free( gfd->request.header );
        gfd->request.header = g_strdup( newh );
        g_free( newh );


        g_strfreev( oh );
        g_strfreev( nh );
        g_free( newheaders );
      } else {
        gaim_debug_misc( __FUNCTION__, "split %s failed\n", newheaders );
      }

      gfd->data_len = 0;
      gfd->len = 0;
      gfd->sentreq = FALSE;
      gfd->startsaving = FALSE;
      gfd->newline = FALSE;

      if (( p = get_header_content( gfd->response.header, "Connection" ))
          != NULL && !strcmp( p, "close" )) {
        /* NTLM will be the death of me */
        int sock;

        gaim_debug_misc( __FUNCTION__, "stupid NTLM closed connection. reopening\n" );

        if (( sock = gaim_proxy_connect( NULL, gfd->website.address, gfd->website.port, url_fetched_cb, gfd )) < 0 ) {
          rvp_async_data( gc, gfd, 0 );
          destroy_fetch_url_data( gfd );
        } else {
          g_free( gfd->response.webdata );
          gfd->response.webdata = NULL;
          gfd->preserve = TRUE;
        }
      } else {
        gfd->resend = TRUE;
      }
      g_free( gfd->response.header );
      gfd->response.header = NULL;
      return;
    }
  }

  p = get_header_content( gfd->response.header, "Content-Type" );
  if ( p && !strcmp( p, "text/xml" )) {
    is_xml = TRUE;
  }

  /* decode the buffer - this has been determined by trial and error
     to be approximately correct, albeit stupid */
  if ( size > 0 ) {
    gchar *converted;
    GError *error = NULL;
    gsize read, written;

    converted = g_convert( gfd->response.webdata, size, "UCS-2LE" /* to */,
                           "UTF-8" /* from */, &read, &written, &error );
    if ( converted ) {
      gsize i;
      gfd->response.webdata = g_realloc( gfd->response.webdata, written / 2 );
      size = written / 2;
      for ( i = 0; i < written; i+= 2 ) {
        gfd->response.webdata[ i / 2 ] = converted[ i ];
      }
      g_free( converted );
    }
  }

  /*
   * if code is unset then we're handling an async message, which has
   * to be carrying an XML payload.
   */
  if ( code == 0 && is_xml ) {
    gint rc;
    /*
     * parse an async notification (e.g. IM, status change)
     */
    rc = parse_notify( gc, gfd );

    /* for now, send a 200 response regardless fixme */
    if ( gfd->sock != -1 ) {
      gchar *reply;

      reply = g_strdup_printf( "HTTP/1.1 %d %s\r\n"
                               "RVP-Notifications-Version: 0.2\r\n\r\n",
                               rc,
                               rc == 200 ? "Successful" : "Parse Error" );
      write( gfd->sock, reply, strlen( reply ));
      g_free( reply );
    }
  } else {
    /* if we're not connected, keep ticking the powerbar FIXME this is
       far too messy */
    if ( !GAIM_CONNECTION_IS_CONNECTED( gc ) &&
         code >= 200 &&
         code < 300
         && rd != NULL &&
         rd->login_step < MAX_LOGIN_STEPS ) {
      rd->login_step++;
      if ( rd->login_step == MAX_LOGIN_STEPS ||
           rd->login_flags == RVP_LOGIN_ALL ) {
        gchar *args[2];

        gaim_connection_set_state( gc, GAIM_CONNECTED );
        serv_finish_login( gc );

        /* send a self-subscribe to update/propchange as well as
           pragma/notify */
        args[0] = ( gchar *)rd->me.buddy;
        args[1] = NULL;
        rvp_send_request( gc, "SUBSCRIBE", args );
      }
    }

    switch ( code ) {
    case 200:
      if ( !strcmp( gfd->method, "SUBSCRIBE" )) {
        /* I should probably double-check, but a 200 response to a
           subscribe means it's a self-subscribe, i.e. a login or
           refresh */
        /* extract the useful info for this subscription */
        gchar *subs_id = get_header_content( gfd->response.header,
                                             "Subscription-Id" );
        gchar *lifetime = get_header_content( gfd->response.header,
                                              "Subscription-Lifetime" );


        rd->login_flags |= RVP_LOGIN_SUBSCRIBE;

        if ( subs_id != NULL ) {
          rd->subs_id = atoi( subs_id );
        }
        if ( lifetime != NULL ) {
          time_t timeout = atol( lifetime );
          if ( timeout == 0 ) {
            gaim_debug_error( __FUNCTION__, "can't parse timeout %s\n",
                              lifetime );
          } else {
            gaim_debug_misc( __FUNCTION__, "main subscription expires in %u seconts\n", timeout );
            timeout = timeout + time( NULL );
            rd->me.expires = timeout;
          }
        }

        if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
          gchar *cmd[2];

          gaim_connection_update_progress( gc, _("Subscribed"), rd->login_step,
                                           MAX_LOGIN_STEPS );

          rvp_send_request( gc, "PROPPATCH", NULL );

          cmd[0] = g_strdup( "d:displayname" );
          cmd[1] = NULL;
          rvp_send_request( gc, "PROPFIND", cmd );
          g_free( cmd[0] );

          rvp_send_request( gc, "SUBSCRIPTIONS", NULL );
        } else {
          /* reset timeout on subscription */
          gaim_debug_misc( __FUNCTION__, "got subs response\n" );
        }
      } else if ( !strcmp( gfd->method, "SUBSCRIPTIONS" )) {
        rd->login_flags |= RVP_LOGIN_SUBSCRIPTIONS;
        if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
          gaim_connection_update_progress( gc, _("Fetched subscriptions"),
                                           rd->login_step, MAX_LOGIN_STEPS );


          /* extract a buddy list from the subscriptions response */
          parse_subscriptions( gc, gfd );

          /* send an ACL self-check */
          rvp_send_request( gc, "ACL", NULL );
        } else {
          /* unlikely... */
          gaim_debug_misc( __FUNCTION__, "got subscriptions response\n" );
        }
      } else if ( !strcmp( gfd->method, "ACL" )) {
        rd->login_flags |= RVP_LOGIN_ACL;
        parse_acls( gc, gfd );
      } else if ( !strcmp( gfd->method, "UNSUBSCRIBE" )) {
        /* only handles self-unsubs xxx */
        rd->me.unsubbed = TRUE;
      }
      break;

    case 207:
      /* has to be XML by definition */
      if ( !strcmp( gfd->method, "PROPFIND" )) {
        rd->login_flags |= RVP_LOGIN_PROPFIND;
      } else if ( !strcmp( gfd->method, "PROPPATCH" )) {
        rd->login_flags |= RVP_LOGIN_PROPPATCH;
      }

      if ( !is_xml ) {
        gaim_connection_error( gc, _("Unparseable reply from server" ));
      } else {
        parse_multistatus( gc, gfd );
      }
      break;

    case 400:
      if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
        gaim_connection_error( gc, _("Bad request"));
        gc->wants_to_die = TRUE;
      } else {
        /* hmm. should be able to recover from this. fixme. */
      }
      break;

    case 403:
      if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
        gaim_connection_error( gc, _("Permission denied"));
        gc->wants_to_die = TRUE;
      } else {
        /* ? */
      }
      break;

    case 404:
      /* treat this as a 412 */
    case 412:
      if ( !strcmp( gfd->method, "NOTIFY" )) {
        gchar *msg = g_strdup_printf( "%s\n%s",
                                      "The following message could not be delivered to all recipients:", "[waider will eventually fix this]" );
        /* serv_got_im( gc, ???, "The following...", flags,
           time(NULL)); */
        /* hacky fixme */
        if ( !g_strstr_len( gfd->request.webdata, gfd->request.length,
                            "\r\nTypingUser: " )) {
          gaim_notify_error( NULL, NULL, msg, NULL );
        }
        g_free( msg );
      } else {
        gaim_debug_error( __FUNCTION__, "%d from %s\n", code, gfd->method );
      }
      break;

    case 500:
      if ( !strcmp( gfd->method, "NOTIFY" )) {
        /* this happens. the message has probably been
           delivered. Microsoft = teh sux0r */
      } else {
        gaim_connection_error( gc, _("Internal server error" ));
        if ( !GAIM_CONNECTION_IS_CONNECTED( gc )) {
          gc->wants_to_die = TRUE;
        }
      }
      break;

    default:
      gaim_debug_misc( __FUNCTION__, "don't know what to do with %d\n", code );
      break;
    }
  }
  /* gfud code will clean up for us */
}

static void rvp_send_request( GaimConnection *gc, gchar *cmd, gchar **args ) {
  RVPData *rd = gc->proto_data;
  GaimFetchUrlData *gfud = NULL;
  xmlChar *content = NULL;
  gint content_length = 0;
  gchar *headers = NULL;
  gchar *url = NULL;
  int sock;

  /* may also need to allow propatch */
  if ( gc->wants_to_die && strcmp( cmd, "UNSUBSCRIBE" )) {
    gaim_debug_misc( __FUNCTION__, "We're dying. Not sending your %s\n", cmd );
    return;
  }

  /* client login/resubscribe */
  if ( !strcmp( cmd, "SUBSCRIBE" )) {
    if ( args == NULL ) {
      url = g_strdup( rd->me.principal );
      if ( rd->subs_id ) {
        headers = g_strdup_printf( "Subscription-Lifetime: %d\r\n"
                                   "Subscription-Id: %d\r\n"
                                   "RVP-From-Principal: %s\r\n",
                                   RVP_LIFETIME,
                                   rd->subs_id,
                                   rd->me.principal );
      } else {
        headers = g_strdup_printf( "Call-Back: %s\r\n"
                                   "Notification-Type: pragma/notify\r\n"
                                   "RVP-From-Principal: %s\r\n"
                                   "Subscription-Lifetime: %d\r\n",
                                   rd->client_url, rd->me.principal,
                                   RVP_LIFETIME );
      }
      /* no content */
    } else {
      GaimBuddy *buddy = (GaimBuddy *)args[0];
      RVPBuddy *rb = buddy->proto_data;

      url = g_strdup( rb->principal );
      if ( rb->subs_id ) {
        headers = g_strdup_printf( "Subscription-Id: %d\r\n"
                                   "Subscription-Lifetime: %d\r\n"
                                   "RVP-From-Principal: %s\r\n",
                                   rb->subs_id,
                                   RVP_LIFETIME,
                                   rd->me.principal );
      } else {
        headers = g_strdup_printf( "Notification-Type: update/propchange\r\n"
                                   "RVP-From-Principal: %s\r\n"
                                   "Subscription-Lifetime: %d\r\n"
                                   "Call-Back: %s\r\n",
                                   rd->me.principal,
                                   RVP_LIFETIME,
                                   rd->me.principal );
      }
    }
  } else if ( !strcmp( cmd, "PROPPATCH" )) {
    xmlDocPtr doc = xmlNewDoc( BAD_CAST "1.0" );
    xmlNodePtr root = xmlNewNode( NULL, BAD_CAST "d:propertyupdate" );
    xmlNodePtr node1, node2, prop;
    gchar *state;

    if ( args != NULL ) {
      state = g_strdup( (gchar *)args[0] );
    } else {
      state = g_strdup( "r:online" ); /* xxx whatever my current state
                                         is */
    }

    xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
              BAD_CAST "r" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
              BAD_CAST "a" );
    xmlDocSetRootElement( doc, root );

    prop = xmlNewNode( NULL, BAD_CAST "r:leased-value" );

    node1 = xmlNewNode( NULL, BAD_CAST "r:value" );
    node2 = xmlNewNode( NULL, BAD_CAST state );
    xmlAddChild( node1, node2 );
    xmlAddChild( prop, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "r:default-value" );
    node2 = xmlNewNode( NULL, BAD_CAST "r:offline" );
    xmlAddChild( node1, node2 );
    xmlAddChild( prop, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "d:timeout" );
    node2 = xmlNewText( BAD_CAST "1200" ); /* xxx timeout value */
    xmlAddChild( node1, node2 );
    xmlAddChild( prop, node1 );

    node1 = xmlNewNode( NULL, BAD_CAST "r:state" );
    xmlAddChild( node1, prop );

    if ( rd->view_id ) {
      gchar *viewid = g_strdup_printf( "%d", rd->view_id );
      prop = xmlNewNode( NULL, BAD_CAST "r:view-id" );
      node2 = xmlNewText( BAD_CAST viewid );
      xmlAddChild( prop, node2 );
      xmlAddChild( node1, prop );
    }

    node2 = xmlNewNode( NULL, BAD_CAST "d:prop" );
    xmlAddChild( node2, node1 );
    node1 = xmlNewNode( NULL, BAD_CAST "d:set" );
    xmlAddChild( node1, node2 );
    xmlAddChild( root, node1 );

    xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
    xmlFreeDoc( doc );

    headers = g_strdup_printf( "Content-Type: text/xml\r\n"
                               "RVP-From-Principal: %s\r\n",
                               rd->me.principal );

    url = g_strdup( rd->me.principal );
  } else if ( !strcmp( cmd, "PROPFIND" )) {
    gint propnum;
    xmlDocPtr doc = xmlNewDoc( BAD_CAST "1.0" );
    xmlNodePtr root = xmlNewNode( NULL, BAD_CAST "d:propfind" );

    xmlNewNs( root, BAD_CAST "DAV:", BAD_CAST "d" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/",
              BAD_CAST "r" );
    xmlNewNs( root, BAD_CAST "http://schemas.microsoft.com/rvp/acl/",
              BAD_CAST "a" );
    xmlDocSetRootElement( doc, root );

    /* practically speaking this is overkill */
    for ( propnum = 0; args[propnum]; propnum++ ) {
      xmlNodePtr prop = xmlNewNode( NULL, BAD_CAST "d:prop" );
      xmlNodePtr propname = xmlNewNode( NULL, BAD_CAST args[propnum] );
      xmlAddChild( prop, propname );
      xmlAddChild( root, prop );
    }

    xmlDocDumpFormatMemory( doc, &content, &content_length, 0 );
    xmlFreeDoc( doc );

    headers = g_strdup_printf( "Depth: 0\r\n" /* always 0 */
                               "Content-Type: text/xml\r\n"
                               "RVP-From-Principal: %s\r\n",
                               rd->me.principal );

    url = g_strdup( rd->me.principal );
  } else if ( !strcmp( cmd, "SUBSCRIPTIONS" )) {
    headers = g_strdup_printf( "Notification-Type: update/propchange\r\n"
                               "RVP-From-Principal: %s\r\n",
                               rd->me.principal );
    url = g_strdup( rd->me.principal );
  } else if ( !strcmp( cmd, "ACL" )) {
    headers = g_strdup_printf( "RVP-From-Principal: %s\r\n",
                               rd->me.principal );
    if ( args != NULL ) {
      /* I think this is wrong */
      /* url = g_strdup( args[0] ); */
      /* and this is right */
      url = g_strdup( rd->me.principal );
      content = (xmlChar *)args[1];
      content_length = (gint)args[2];
    } else {
      url = g_strdup( rd->me.principal );
    }
  } else if ( !strcmp( cmd, "NOTIFY" )) {
    headers = g_strdup_printf( "RVP-From-Principal: %s\r\n"
                               "Content-Type: text/xml\r\n"
                               "RVP-Ack-Type: DeepOr\r\n",
                               rd->me.principal );

    url = (gchar *)args[0];
    content = (xmlChar *)args[1];
    content_length = (gint)args[2];
  } else if ( !strcmp( cmd, "UNSUBSCRIBE" )) {
    if ( args == NULL ) {
      /* self-unsubscribe */
      url = g_strdup( rd->me.principal );
      headers = g_strdup_printf( "RVP-From-Principal: %s\r\n"
                                 "Subscription-Id: %d\r\n",
                                 rd->me.principal,
                                 rd->subs_id );
    } else {
      GaimBuddy *buddy = (GaimBuddy *)args[0];
      RVPBuddy *rbuddy = buddy->proto_data;
      if ( rbuddy == NULL ||
           rbuddy->subs_id == 0 ||
           rbuddy->principal == NULL ) {
        gaim_debug_error( __FUNCTION__, "Buddy problem unsubbing\n" ); /* Xxx */
        rvp_dump_buddy( buddy );
        return;
      }

      url = g_strdup( rbuddy->principal );
      headers = g_strdup_printf( "RVP-From-Principal: %s\r\n"
                                 "Subscription-Id: %d\r\n",
                                 rbuddy->principal,
                                 rbuddy->subs_id );
    }
  }

  if ( rd->auth_type == RVP_AUTH_NTLM ) {
    gaim_debug_misc( __FUNCTION__, "Preemptively adding NTLM\n" );
    gchar *ntlm = get_ntlm_msg1( rd->domain, rd->client_host );
    gchar *newheaders = g_strdup_printf( "Authorization: %s\r\n%s", ntlm, headers );
    g_free( headers );
    headers = newheaders;
  }

  gfud = g_new0( GaimFetchUrlData, 1 );
  gfud->request.header = headers;
  gfud->request.webdata = (gchar *)content; /* xxx */
  gfud->request.length = content_length;
  gfud->method = g_strdup( cmd );
  gfud->callback = rvp_async_data;
  gfud->user_data = gc;
  gfud->url = url;
  gaim_url_parse( gfud->url, &gfud->website.address, &gfud->website.port,
                  &gfud->website.page, &gfud->website.user,
                  &gfud->website.passwd );

  if (( sock =
        gaim_proxy_connect( NULL, gfud->website.address,
                            gfud->website.port, url_fetched_cb, gfud )) < 0 ) {
    rvp_async_data( gc, gfud, 0 ); /* xxx */
    destroy_fetch_url_data( gfud );
  }
}

/*
 * attached to the listening port
 */
static void rvp_listener_callback( gpointer data, gint source,
                                   GaimInputCondition cond ) {
  GaimConnection *gc = data;
  GaimFetchUrlData *fetch = NULL;
  RVPData *rd = gc->proto_data;
  int sd;

  /* this might have happened because of a disconnection... */
  if ( rd == NULL ) {
    gaim_debug_misc( __FUNCTION__, "erk. rd is null\n" );
    return;
  }

  if (( sd = accept( rd->ld, 0, 0 )) < 0 ) {
      perror( "Accept failed.\n" );
      /* fixme drop the listener and make a new one? */
      gc->wants_to_die = TRUE;
      return;
  }

  /* These are incoming connections that we only expect to read from */
  fetch = g_new0( GaimFetchUrlData, 1 );
  fetch->callback = rvp_async_data;
  fetch->user_data = gc;
  fetch->sentreq = TRUE;
  fetch->data_len = 4096;
  fetch->response.webdata = g_malloc( fetch->data_len );
  fetch->sock = sd;
  fetch->inpa = gaim_input_add( sd, GAIM_INPUT_READ, url_fetched_cb, fetch );

  return;
}

/*
 * This is invoked via Gaim's event loop from rvp_login
 */
static void rvp_login_connect( gpointer data, gint source,
                               GaimInputCondition cond ) {
  GaimConnection *gc = data;
  struct rvp_data *rd = gc->proto_data;
  char buf[RVP_BUF_SIZE];
  char hostname[1024];
  char *tmp, *tmp2;
  struct hostent *p;
  int myhostname_found=FALSE;

  gaim_debug_misc( __FUNCTION__,  "enter\n" );

  if ( rd->fd != source ) {
    gaim_debug_misc( __FUNCTION__, "source fd was %d, now %d\n",
                     rd->fd, source );
    rd->fd = source;
  }

  if ( rd->fd == -1 ) {
    gaim_connection_error(gc, _("Unable to connect"));
    gc->wants_to_die = TRUE;
    return;
  }

  /*
   * Set up the callback socket
   */
  rvp_setup_listener( gc );

  /* check for failure - fixme, have rvp_setup_listener return a
     useful rc */
  if ( gc->wants_to_die == TRUE ) {
    return;
  }

  /* connect the listener up to Gaim's event loop */
  gaim_debug_misc( __FUNCTION__, "listener hookup\n" );
  rd->linpa =
    gaim_input_add( rd->ld, GAIM_INPUT_READ, rvp_listener_callback, gc );

  /* Create a session ID */
  rd->session_id = rvp_get_sessid();

  /* get server and client info */
  /* FIXME why bother doing this? */
  strcpy( hostname, gaim_connection_get_account( gc )->username );
  tmp = strtok(hostname, "@");
  rd->user = g_strdup( tmp );

  strcpy(hostname, gaim_connection_get_account( gc )->username);

  if ( gethostname(hostname, sizeof(hostname) -1) != 0 ) {
    perror("gethostname() failure");
    abort();
  }

  hostname[sizeof(hostname) - 1] = 0;
  p = gethostbyname(hostname);
  if ( p == NULL ) {
    perror("gethostname() failure");
    abort();
  }

  /*
   * Some systems pick a hostname which isn't in DNS.  So, we
   * we'll try to use the IP address as the most general solution,
   * as long as the IP adress does not start with "127."
   *
   * FIXME this is also pretty bogus, I think.
   */
  if ( p && p->h_addr_list && p->h_addr_list[0] ) {
    struct in_addr addr;

    addr = *(struct in_addr *)(p->h_addr_list[0]);

    strncpy(buf,inet_ntoa(addr),sizeof(buf));
    if ( strncmp(buf,"127.",4) != 0 ) {
      myhostname_found = TRUE;
      strncpy(hostname,buf,sizeof(hostname));
      rd->client_host = g_strdup(hostname);
    }
  }

  /* Does hostname contain a period */
  if ( !myhostname_found && strchr(p->h_name, '.')) {
    /* Yes, use it */
    strncpy(hostname,p->h_name,sizeof(hostname));
    rd->client_host = g_strdup(tmp);

    myhostname_found = TRUE;

  } else if ( !myhostname_found) {
    /* No, search all aliases */
    int x;

    if (p->h_aliases) {
      for (x = 0; p->h_aliases[x]; ++x) {
        if (p->h_aliases[x] && strchr(p->h_aliases[x], '.') &&
            (!strncasecmp(p->h_aliases[x], p->h_name, strlen(p->h_name)))) {
          strncpy(hostname,p->h_aliases[x],sizeof(hostname));
          myhostname_found = TRUE;
          rd->client_host = g_strdup(hostname);

          break;
        }
      }
    }
  }
  hostname[sizeof(hostname) - 1] = 0;

  /* If all else fails, use GAIM_HOSTNAME environment variable */
  /* might be smarter to do this FIRST and then skip the rest... */
  if ( (tmp2=getenv("GAIM_HOSTNAME")) != NULL  ) {
    strncpy(hostname,tmp2,sizeof(hostname));
    rd->client_host = g_strdup(hostname);
    myhostname_found = TRUE;
  }

  /* XXX */
  if ( ! myhostname_found ) {
    fprintf(stderr, "Unable to find our fully qualified hostname or IP address.  Sorry, I'm giving up.\n");
    abort();
  } else {
    gaim_debug_misc( __FUNCTION__, "determined my hostname as %s\n", hostname );
  }

  rd->client_url = g_strdup_printf( "http://%s:%d", hostname, rd->lport );

  /*
   * push a request out
   */
  rvp_send_request( gc, "SUBSCRIBE", NULL );
  gaim_debug_misc( __FUNCTION__, "exit\n" );
}

/*
 * Initial entrypoint for establishing a connection
 *
 * invokes rvp_login_connect via Gaim's event loop.
 */
static void rvp_login( GaimAccount *account ) {
  GaimConnection *gc = gaim_account_get_connection( account );
  RVPData *rd;
  const char *host = NULL, *domain, *id;
  const char *username;
  int port;
  gchar **split;
  gchar *msg;

  if ( gc->proto_data != NULL ) {
    g_free( gc->proto_data ); /* xxx */
  }
  gc->proto_data = g_new0( RVPData, 1 );
  rd = gc->proto_data;

  rd->login_step = 0;
  rd->me.buddy = g_new0( GaimBuddy, 1 );
  rd->me.buddy->proto_data = rd;

  username = gaim_account_get_username( account );
  split = g_strsplit( username, "@", 2 );

  if ( split == NULL ) {
    gc->wants_to_die = TRUE;
    gaim_connection_error( gc, _("Invalid username."));
    return;
  }

  if ( split[1] != NULL ) {
    gchar *srvname = g_strconcat( "_rvp._tcp.", split[1], NULL );

    msg = g_strdup_printf( _("Locating RVP server for %s"), split[1] );
    gaim_debug_misc( __FUNCTION__, "%s\n", msg );
    g_free( msg );

    rd->me.service = gethostbysrv( srvname, rd->me.service );
    if ( rd->me.service->host != NULL ) {
      host = rd->me.service->host;
      rd->port = rd->me.service->port;
    }
    g_free( srvname );

    /* if host == NULL:
       gethostbyname( split[1] );
       nmblookup( split[1] );
    */
  } else {
    host = gaim_account_get_string( account, "host", NULL );
    rd->port = 80; /* default */
  }

  if ( rd->port != 80 ) {
    rd->me.principal = g_strdup_printf( "http://%s:%d/instmsg/aliases/%s",
                                         host, rd->port, split[0] );
  } else {
    rd->me.principal = g_strdup_printf( "http://%s/instmsg/aliases/%s",
                                         host, split[0] );
  }
  rd->me.buddy->name = g_strconcat( split[0], "@", host, NULL );

  g_strfreev( split );

  if ( host == NULL ) {
    gc->wants_to_die = TRUE;
    gaim_connection_error( gc, _("No login host specified."));
    return;
  }

  msg = g_strdup_printf( _("Connecting to %s:%d"), host, rd->port );
  gaim_connection_update_progress( gc, msg, rd->login_step, MAX_LOGIN_STEPS );
  gaim_debug_misc( __FUNCTION__, "%s\n", msg );
  g_free( msg );

  domain = gaim_account_get_string( account, "domain", NULL );
  id = gaim_account_get_string( account, "id", NULL );
  port = gaim_account_get_int( account, "port", -1 );

  rd->host = g_strdup( host );

  rd->domain = g_strdup( domain );
  rd->authid = g_strdup( id );
  rd->listenport = port;

  rd->fd =
    gaim_proxy_connect( account, rd->host, rd->port, rvp_login_connect, gc );

  if ( rd->fd < 0 ) {
    gaim_connection_error(gc, _("Failed to connect to server."));
  }
}

static gboolean rvp_load( GaimPlugin *plugin ) {
  gaim_debug_misc( __FUNCTION__, "rvp_load\n" );
  return TRUE;
}

static gboolean rvp_unload( GaimPlugin *plugin ) {
  gaim_debug_misc( __FUNCTION__,  "rvp_unload\n" );
  return TRUE;
}

static GaimPluginPrefFrame * get_plugin_pref_frame(GaimPlugin *plugin) {
  GaimPluginPrefFrame *frame;
  GaimPluginPref *ppref;

  frame = gaim_plugin_pref_frame_new();

  ppref = gaim_plugin_pref_new_with_label(_("Conversations"));
  gaim_plugin_pref_frame_add(frame, ppref);

  ppref = gaim_plugin_pref_new_with_name_and_label(
                                                   "/plugins/prpl/rvp/conv_close_notice",
                                                   _("Display conversation closed notices"));
  gaim_plugin_pref_frame_add(frame, ppref);

  ppref = gaim_plugin_pref_new_with_name_and_label(
                                                   "/plugins/prpl/rvp/conv_timeout_notice",
                                                   _("Display timeout notices"));
  gaim_plugin_pref_frame_add(frame, ppref);

  return frame;
}

static GaimPluginUiInfo prefs_info = {
  get_plugin_pref_frame
};

static GaimPluginProtocolInfo prpl_info = {
  0, /* no options */
  NULL,                   /* user_splits */
  NULL,                   /* protocol_options */
  { "png", 0, 0, 96, 96, GAIM_ICON_SCALE_SEND },    /* icon_spec */
  rvp_list_icon,          /* list_icon */
  rvp_list_emblems,       /* list_emblems */
  rvp_status_text,        /* status_text */
  NULL /* rvp_tooltip_text */,       /* tooltip_text */
  rvp_away_states,        /* away_states */
  rvp_buddy_menu,         /* blist_node_menu */
  NULL,                   /* chat_info */
  NULL,                   /* chat_info_defaults */

  rvp_login,              /* login */
  rvp_close,              /* close */
  rvp_send_im,            /* send_im */
  NULL,                   /* set_info */
  rvp_send_typing,        /* send_typing */
  NULL /* rvp_get_info */,           /* get_info */
  rvp_set_away,           /* set_away */
  rvp_set_idle,           /* set_idle */
  NULL,                   /* change_passwd */
  rvp_add_buddy,          /* add_buddy */
  NULL,                   /* add_buddies */
  rvp_rem_buddy,          /* remove_buddy */
  NULL,                   /* remove_buddies */
  rvp_add_permit,         /* add_permit */
  rvp_add_deny,           /* add_deny */
  rvp_rem_permit,         /* rem_permit */
  rvp_rem_deny,           /* rem_deny */
  rvp_set_permit_deny,    /* set_permit_deny */
  NULL,                   /* warn */
  rvp_chat_join,          /* join_chat */
  NULL,                   /* reject chat invite */
  NULL,                   /* get_chat_name */
  rvp_chat_invite,        /* chat_invite */
  rvp_chat_leave,         /* chat_leave */
  NULL,                   /* chat_whisper */
  rvp_chat_send,          /* chat_send */
  rvp_keepalive,          /* keepalive */
  NULL,                   /* register_user */
  NULL,                   /* get_cb_info */
  NULL,                   /* get_cb_away */
  NULL,                   /* alias_buddy */
  NULL /* rvp_group_buddy */,        /* group_buddy */
  NULL /* rvp_rename_group */,       /* rename_group */
  NULL,                   /* buddy_free */
  NULL /* rvp_convo_closed */,       /* convo_closed */
  rvp_normalize,          /* normalize */
  NULL /* rvp_set_buddy_icon */,     /* set_buddy_icon */
  NULL /* rvp_remove_group */,       /* remove_group */
  NULL,                   /* get_cb_real_name */
  NULL,                   /* set_chat_topic */
  NULL,                   /* find_blist_chat */
  NULL,                   /* roomlist_get_list */
  NULL,                   /* roomlist_cancel */
  NULL,                   /* roomlist_expand_category */
  NULL /* rvp_can_receive_file */,   /* can_receive_file */
  NULL /* rvp_send_file */          /* send_file */
};

static GaimPluginInfo info = {
  GAIM_PLUGIN_MAGIC,
  GAIM_MAJOR_VERSION,
  GAIM_MINOR_VERSION,
  GAIM_PLUGIN_PROTOCOL,                             /**< type           */
  NULL,                                             /**< ui_requirement */
  0,                                                /**< flags          */
  NULL,                                             /**< dependencies   */
  GAIM_PRIORITY_DEFAULT,                            /**< priority       */

  PROTO_RVP,                                        /**< id             */
  "RVP",                                            /**< name           */
  VERSION,                                          /**< version        */
                                                      /**  summary        */
  N_("RVP Protocol Plugin"),
  /**  description    */
  N_("RVP Protocol Plugin"),
  "Weihua Sun <weihua@lucent.com>, Waider <waider@waider.ie>", /**< author         */
  GAIM_WEBSITE,                                     /**< homepage       */

  rvp_load,                                         /**< load           */
  rvp_unload,                                       /**< unload         */
  NULL,                                             /**< destroy        */

  NULL,                                             /**< ui_info        */
  &prpl_info,                                       /**< extra_info     */
  &prefs_info,                                      /**< prefs_info     */
  NULL /* rvp_actions */
};

static void init_plugin( GaimPlugin *plugin ) {
  GaimAccountOption *option;
  gchar *authid = NULL;
  gchar *fixedport = NULL;
  gint port = 0;

  option = gaim_account_option_string_new(_("Auth Host"), "host", "" );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  option = gaim_account_option_string_new(_("Auth Domain"), "domain", "" );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  /* compatability with old code */
  authid = getenv( "RVP_AUTH_ID" );
  fixedport = getenv( "RVP_FIXED_PORT" );
  if ( fixedport != NULL ) {
    port = atoi( fixedport );
  }

  option = gaim_account_option_string_new(_("Auth ID"), "id", authid );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  option = gaim_account_option_int_new(_("Fixed Port"), "port", port );
  prpl_info.protocol_options = g_list_append( prpl_info.protocol_options,
                                              option );

  /* ???
     gaim_prefs_add_none( "/plugins/prpl/rvp");
     gaim_prefs_add_bool("/plugins/prpl/rvp/conv_close_notice",   TRUE);
     gaim_prefs_add_bool("/plugins/prpl/rvp/conv_timeout_notice", TRUE);
  */
}

GAIM_INIT_PLUGIN(rvp, init_plugin, info);
